import os

class Worker:
    COST = int(os.environ.get("WORKER_COST", "20"))
    HP = int(os.environ.get("WORKER_HP", "10"))
    SPEED = float(os.environ.get("WORKER_SPEED", "3"))
    TRIPS = int(os.environ.get("WORKER_TRIPS", "3"))
    WORK_RATE = float(os.environ.get("WORK_RATE", "1"))
    ENCUMBERED_RATE = float(os.environ.get("WORKER_ENCUMBERED_RATE", "0.3"))

class Fighter:
    COST = int(os.environ.get("FIGHTER_COST", "40"))
    HP = int(os.environ.get("FIGHTER_HP", "15"))
    SPEED = float(os.environ.get("FIGHTER_SPEED", "2"))
    ATTACK = float(os.environ.get("FIGHTER_ATTACK", "3.5"))
    RANGE = float(os.environ.get("FIGHTER_RANGE", "1.5"))
    NUM_ATTACKS = int(os.environ.get("FIGHTER_NUM_ATTACKS", "2"))
    LIFESPAN = int(os.environ.get("FIGHTER_LIFESPAN", "30"))

class Settler:
    COST = int(os.environ.get("SETTLER_COST", "30"))
    HP = int(os.environ.get("SETTLER_HP", "10"))
    SPEED = float(os.environ.get("SETTLER_SPEED", "1.5"))
    LIFESPAN = int(os.environ.get("SETTLER_LIFESPAN", "40"))



import os.path
from os import getcwd
import random
from codequest22.server.events import *
from codequest22.server.network import NetworkManager
from codequest22.server.replay import ReplayManager
from codequest22.server.requests import *
from codequest22.server.global_map import GlobalMap, UnreachableException
from codequest22.server.ant import AntTypes
from codequest22 import stats
from traceback import format_exc
from pyqtree import Index
import numpy as np



def start_server(map_path, replay_path, recv_queue, visual_queue, error_queue, client_queues, is_visual):
    try:
        GlobalMap.load_map(map_path)
        spindex = Index(bbox=(-0.5, -0.5, GlobalMap.h - 0.5, GlobalMap.w - 0.5))
        ReplayManager.set_output(replay_path)
        NetworkManager.set_queues(visual_queue, recv_queue, client_queues, is_visual)

        # Send initial map data.
        NetworkManager.send_internal_obj({
            "type": "map", 
            "obj": GlobalMap.map_obj,
            "energy_info": GlobalMap.energy_tiles,
            "zones": GlobalMap.capture_zones,
        })

        zone_activated = [False for _ in GlobalMap.capture_zones]

        start_up = stats.hill.GRACE_PERIOD
        zone_times = len(GlobalMap.capture_zones) * stats.hill.NUM_ACTIVATIONS
        zone_allocations = [x % len(GlobalMap.capture_zones) for x in range(zone_times)]
        random.shuffle(zone_allocations)
        # Each capture zone will light up the same number of times.
        # Each time, this will take somewhere between 60 and 80 ticks, with 20 - 30 tick cooldown
        min_zone_time = stats.hill.MIN_ZONE_TIME
        max_zone_time = stats.hill.MAX_ZONE_TIME
        min_wait_time = stats.hill.MIN_WAIT_TIME
        max_bar = stats.general.SIMULATION_TICKS - start_up - zone_times * (max_zone_time + min_wait_time)
        zone_spots = [random.randint(1, max_bar) for _ in range(zone_times)]
        zone_spots.sort()
        zone_events = []
        cur_addition = start_up
        for i, v in enumerate(zone_spots):
            actual_zone_time = random.randint(min_zone_time, max_zone_time)
            zone_events.append((v + cur_addition, actual_zone_time, zone_allocations[i]))
            zone_events.append((v + cur_addition + actual_zone_time, None, zone_allocations[i]))
            cur_addition += actual_zone_time + min_wait_time

        tile_activated = [1 for _ in GlobalMap.energy_tiles]
        tile_start_up = stats.energy.GRACE_PERIOD
        min_tile_wait_time = stats.energy.MIN_WAIT_TIME
        min_tile_activation_time = stats.energy.MIN_OVERCHARGE_TIME
        max_tile_activation_time = stats.energy.MAX_OVERCHARGE_TIME
        n_times = stats.energy.NUM_ACTIVATIONS
        tile_times = [
            [
                random.randint(1, stats.general.SIMULATION_TICKS - tile_start_up - (min_tile_wait_time + max_tile_activation_time) * n_times)
                for _ in range(n_times)
            ] for _ in range(len(tile_activated))
        ]
        tile_events = []
        for x in range(len(tile_times)):
            tile_times[x].sort()
            for y in range(n_times):
                start = tile_times[x][y] + tile_start_up + y * (min_tile_wait_time + max_tile_activation_time)
                length = random.randint(min_tile_activation_time, max_tile_activation_time)
                tile_events.append((start, length, x))
                tile_events.append((start+length, None, x))

        player_data = NetworkManager.recv_client_response()
        # Send every client their index.
        for i in range(len(client_queues)):
            NetworkManager.send_client(i, {
                "index": i,
                "total": len(client_queues),
            })

        NetworkManager.broadcast_obj({
            "type": "map", 
            "obj": GlobalMap.map_obj,
            "energy_tiles": {
                tuple(e.position[::-1]): e.amount
                for e in GlobalMap.energy_tiles
            }
        })

        ants = [
            {} for d in player_data
        ]
        player_energy = [stats.general.STARTING_ENERGY for d in player_data]
        player_health = [stats.general.QUEEN_HEALTH for d in player_data]
        defeated = [False for d in player_data]
        hill_score = [0 for _ in player_data]
        # Remove cwd from absolute paths
        image_names = [os.path.relpath(d["image"], getcwd()) for d in player_data]
        NetworkManager.send_internal_obj({
            "type": "player_data", 
            "names": [d["name"] for d in player_data], 
            "images": image_names,
            "energy": player_energy,
            "hill": hill_score,
            "health": player_health,
        })

        ant_ids = [0 for d in player_data]
        alive_ants = [
            {
                AntTypes.WORKER: 0,
                AntTypes.FIGHTER: 0,
                AntTypes.SETTLER: 0,
            } 
            for d in player_data
        ]

        player_events = []
        server_events = []
        failed_requests = []
        cur_tick = 0
        last_succeeded_request = 0
        while True:
            NetworkManager.broadcast_obj({
                "type": "tick", 
                "events": player_events,
                "failed_requests": failed_requests,
            })
            player_events = []
            server_events = []
            failed_requests = []
            group_requests = NetworkManager.recv_client_response()
            for i, d in enumerate(player_data):
                n_spawns = 0
                if defeated[i] and group_requests[i]:
                    group_requests[i][0].player_index = i
                    group_requests[i][0].reason = f"You are defeated. Please stop sending events."
                    failed_requests.append(group_requests[i][0])
                    continue
                for req in group_requests[i]:
                    if isinstance(req, SpawnRequest):
                        if n_spawns >= stats.general.MAX_SPAWNS_PER_TICK:
                            req.player_index = i
                            req.reason = f"You can only spawn {stats.general.MAX_SPAWNS_PER_TICK} ants per tick."
                            failed_requests.append(req)
                            continue
                        if player_energy[i] < req.cost:
                            req.player_index = i
                            req.reason = f"Does not have enough energy to request. Requested spawn with cost {req.cost}, but you have {player_energy[i]} energy"
                            failed_requests.append(req)
                            continue
                        if req.cost < AntTypes.get_stats(req.ant_type).COST:
                            req.player_index = i
                            req.reason = f"The basic ant of this type requires at least {AntTypes.get_stats(req.ant_type).COST} energy, you cannot pay {req.cost} energy"
                            failed_requests.append(req)
                            continue
                        if alive_ants[i][AntTypes.WORKER] + alive_ants[i][AntTypes.FIGHTER] + alive_ants[i][AntTypes.SETTLER] >= stats.general.MAX_ANTS_PER_PLAYER:
                            req.player_index = i
                            req.reason = f"You already have {stats.general.MAX_ANTS_PER_PLAYER} on the field, cannot spawn more until ants die."
                            failed_requests.append(req)
                            continue
                        if req.id in ants[i]:
                            req.player_index = i
                            req.reason = f"There already exists one of your ants with id {req.id}"
                            failed_requests.append(req)
                            continue
                        if req.id is None:
                            req.id = f"ant-{ant_ids[i]}"
                            ant_ids[i] += 1
                        klass = AntTypes.get_class(req.ant_type)
                        ant_obj = klass(i, req.id, GlobalMap.player_spawns[i], req.cost, req.color or (0, 0, 0, 0))
                        if req.goal is not None:
                            try:
                                ant_obj.set_goal(req.goal)
                            except UnreachableException:
                                req.player_index = i
                                req.reason = f"Ant goal position is unreachable. Ant position: {ant_obj.position[::-1]}, Ant goal: {req.goal[::-1]}"
                                failed_requests.append(req)
                                continue
                        ants[i][req.id] = ant_obj
                        spindex.insert((ant_obj.player_index, ant_obj.id), (*ant_obj.position, *ant_obj.position))
                        player_energy[i] -= req.cost
                        player_events.append(SpawnEvent(ant_obj, req.cost))
                        server_spawn = SpawnEvent(ant_obj, req.cost, ant_obj.color, getattr(ant_obj, "_follow_path", []), ant_obj.goal)
                        server_events.append(server_spawn)
                        alive_ants[i][req.ant_type] += 1
                        n_spawns += 1
                        last_succeeded_request = cur_tick
                    elif isinstance(req, GoalRequest):
                        if req.ant_id not in ants[i]:
                            req.player_index = i
                            req.reason = f"This ant does not exist: {req.ant_id}"
                            failed_requests.append(req)
                            continue
                        x, y = req.position
                        try:
                            ants[i][req.ant_id].set_goal((y, x))
                            last_succeeded_request = cur_tick
                        except UnreachableException:
                            req.player_index = i
                            req.reason = f"Ant goal position is unreachable. Ant position: {ant_obj.position[::-1]}, Ant goal: {(x, y)}"
                            failed_requests.append(req)
                            continue
            for i in range(len(player_data)):
                for ant in ants[i].values():
                    if ant.alive():
                        ant.tick()
            for i in range(len(player_data)):
                for k, v in ants[i].items():
                    if v.alive() and v.moved:
                        spindex.remove((v.player_index, v.id), (*v.old_position, *v.old_position))
                        spindex.insert((v.player_index, v.id), (*v.position, *v.position))
                        player_events.append(MoveEvent(v))
                        server_move = MoveEvent(v, getattr(v, "_follow_path", []))
                        server_events.append(server_move)
            # Handle all attacks
            for i in range(len(player_data)):
                for k, v in ants[i].items():
                    if v.TYPE == AntTypes.FIGHTER:
                        # Search in a range around this ant.
                        count = 0
                        attack_range = v.range
                        possible = spindex.intersect((
                            v.position[0]-attack_range,
                            v.position[1]-attack_range,
                            v.position[0]+attack_range,
                            v.position[1]+attack_range,
                        ))
                        to_remove = []
                        for i2 in range(len(possible)):
                            a = ants[possible[i2][0]][possible[i2][1]]
                            if (
                                (a.position[0] - v.position[0]) * (a.position[0] - v.position[0]) + 
                                (a.position[1] - v.position[1]) * (a.position[1] - v.position[1])
                            ) > attack_range * attack_range:
                                to_remove.append(i2)
                            elif possible[i2][0] == i:
                                to_remove.append(i2)
                        for i2 in to_remove[::-1]:
                            del possible[i2]
                        # We now have all possible candidates for attack.
                        n_attacks = min(len(possible), v.num_attacks)
                        choices = np.random.choice(len(possible), size=n_attacks, replace=False)
                        for c in choices:
                            i2, id2 = possible[c]
                            # Attack!
                            v.attack(ants[i2][id2])
                            player_events.append(AttackEvent(v, ants[i2][id2]))
                            server_events.append(AttackEvent(v, ants[i2][id2]))
                            count += 1
                        if count < v.num_attacks:
                            for i2 in range(len(player_data)):
                                if i != i2:
                                    dx = GlobalMap.player_spawns[i2][0] - v.position[0]
                                    dy = GlobalMap.player_spawns[i2][1] - v.position[1]
                                    if dx * dx + dy * dy <= attack_range * attack_range:
                                        # Attack Queen
                                        cached_health = player_health[i2]
                                        player_health[i2] -= v.attack_damage * (v.num_attacks - count)
                                        player_health[i2] = max(player_health[i2], 0)
                                        ev = QueenAttackEvent(v, i2, player_health[i2])
                                        player_events.append(ev)
                                        server_events.append(ev)
                                        if player_health[i2] <= 0 and cached_health > 0:
                                            defeated[i2] = True
                                            hill_score[i] += hill_score[i2]
                                            hill_score[i2] = 0
                                            ev2 = TeamDefeatedEvent(i2, i, hill_score[i])
                                            player_events.append(ev2)
                                            server_events.append(ev2)
            # Handle deaths
            for i in range(len(player_data)):
                to_remove = []
                for k, v in ants[i].items():
                    if not v.alive():
                        to_remove.append(k)
                        player_events.append(DieEvent(v))
                        server_events.append(DieEvent(v))
                for k in to_remove:
                    spindex.remove((ants[i][k].player_index, ants[i][k].id), (*ants[i][k].position, *ants[i][k].position))
                    alive_ants[i][ants[i][k].TYPE] -= 1
                    del ants[i][k]
            # Worker actions
            # step 1: Queen depositing
            for i in range(len(player_data)):
                for k, v in ants[i].items():
                    if v.TYPE == AntTypes.WORKER and v.encumbered:
                        # Am I close to my queen?
                        ipos = (round(v.position[0]), round(v.position[1]))
                        if GlobalMap.player_spawns[i][0] == ipos[0] and GlobalMap.player_spawns[i][1] == ipos[1]:
                            # Deposit my energy
                            dv = DepositEvent(v, player_energy[i])
                            player_energy[i] = dv.total_energy
                            player_events.append(dv)
                            server_events.append(dv)
                            v.free_energy()
            # step 2: Tile production
            for t in GlobalMap.energy_tiles:
                if t.cur_delay > 0:
                    t.cur_delay -= 1
            ballots = [[] for _ in GlobalMap.energy_tiles]
            for i in range(len(player_data)):
                for k, v in ants[i].items():
                    if v.TYPE == AntTypes.WORKER and not v.encumbered:
                        # Am I close to a food tile?
                        for i2, tile in enumerate(GlobalMap.energy_tiles):
                            ipos = (round(v.position[0]), round(v.position[1]))
                            if tile.position == ipos:
                                # Pick up some energy
                                ballots[i2].append((i, k))
            for i2, (tile, b) in enumerate(zip(GlobalMap.energy_tiles, ballots)):
                mult = tile_activated[i2]
                if tile.cur_delay > 0: continue
                for x in range(tile.per_tick):
                    if not b: continue
                    tile.cur_delay = tile.max_delay
                    i, k = random.choice(b)
                    v = ants[i][k]
                    v.encumbered = True
                    v.encumbered_energy = v.work_rate * tile.amount * mult
                    player_events.append(ProductionEvent(v))
                    server_events.append(ProductionEvent(v))
                    for b2 in ballots:
                        if (i, k) in b2:
                            b2.remove((i, k))
            # Settler scoring
            for i in range(len(player_data)):
                for k, v in ants[i].items():
                    if v.TYPE == AntTypes.SETTLER:
                        rounded_pos = round(v.position[0]), round(v.position[1])
                        for i2, zone in enumerate(zone_activated):
                            if not zone: continue
                            if rounded_pos in GlobalMap.capture_zones[i2]:
                                hill_score[i] += 1
                                server_events.append(SettlerScoreEvent(v, 1))
                                player_events.append(SettlerScoreEvent(v, 1))
                                break
            # Zone activation / deactivation
            for tick, length, z in zone_events:
                if cur_tick == tick:
                    if length is None:
                        zone_activated[z] = False
                        server_events.append(ZoneDeactivateEvent(z, [p[::-1] for p in GlobalMap.capture_zones[z]]))
                        player_events.append(ZoneDeactivateEvent(z, [p[::-1] for p in GlobalMap.capture_zones[z]]))
                    else:
                        zone_activated[z] = True
                        server_events.append(ZoneActiveEvent(z, length, [p[::-1] for p in GlobalMap.capture_zones[z]]))
                        player_events.append(ZoneActiveEvent(z, length, [p[::-1] for p in GlobalMap.capture_zones[z]]))
            # Tile activation / deactivation
            for tick, length, z in tile_events:
                if cur_tick == tick:
                    if length is None:
                        tile_activated[z] = 1
                        server_events.append(FoodTileDeactivateEvent(GlobalMap.energy_tiles[z].position[::-1]))
                        player_events.append(FoodTileDeactivateEvent(GlobalMap.energy_tiles[z].position[::-1]))
                    else:
                        tile_activated[z] = 2
                        server_events.append(FoodTileActiveEvent(GlobalMap.energy_tiles[z].position[::-1], length, 2))
                        player_events.append(FoodTileActiveEvent(GlobalMap.energy_tiles[z].position[::-1], length, 2))
            NetworkManager.send_internal_obj({
                "type": "tick",
                "events": server_events,
                "failed_requests": failed_requests,
            })
            NetworkManager.wait_visual_response()
            cur_tick += 1
            # if cur_tick % 10 == 0:
            #     print(cur_tick)
            # Check for any players with no ants and not enough nectar to buy any ants
            for i in range(len(player_data)):
                if not defeated[i] and len(ants[i].keys()) == 0:
                    if player_energy[i] < min(stats.ants.Fighter.COST, stats.ants.Worker.COST, stats.ants.Settler.COST):
                        defeated[i] = True
            if cur_tick == stats.general.SIMULATION_TICKS or cur_tick - last_succeeded_request > 100:
                NetworkManager.broadcast_obj({"type": "finish"})
                for i in range(len(hill_score)):
                    if player_health[i] <= 0:
                        hill_score[i] = 0
                sorted_hill_score = list(sorted(hill_score))
                cur_score = 100
                score_map = {}
                for v in sorted_hill_score[::-1]:
                    if v not in score_map:
                        score_map[v] = cur_score
                        cur_score -= 25
                NetworkManager.send_internal_obj({
                    "type": "winner",
                    "indicies": [i for i in range(len(hill_score)) if hill_score[i] == max(hill_score)],
                    "score": [25 if player_health[i] <= 0 else score_map[hill_score[i]] for i in range(len(hill_score))]
                })
                ReplayManager.close()
                if not is_visual:
                    error_queue.put("server")
                return
            if len(defeated) - sum(defeated) <= 1 and (len(defeated) != 1 or len(defeated) == sum(defeated)):
                if sum(defeated) == len(defeated) - 1 and max(hill_score) != max(hill_score[i] for i in range(len(hill_score)) if not defeated[i]):
                    # Keep going
                    pass
                else:
                    # Game over.
                    for i in range(len(hill_score)):
                        if player_health[i] <= 0:
                            hill_score[i] = 0
                    sorted_hill_score = list(sorted(hill_score))
                    cur_score = 100
                    score_map = {}
                    for v in sorted_hill_score[::-1]:
                        if v not in score_map:
                            score_map[v] = cur_score
                            cur_score -= 25
                    NetworkManager.send_internal_obj({
                        "type": "winner",
                        "indicies": [i for i in range(len(hill_score)) if hill_score[i] == max(hill_score)],
                        "score": [25 if player_health[i] <= 0 else score_map[hill_score[i]] for i in range(len(hill_score))]
                    })
                    ReplayManager.close()
                    if not is_visual:
                        error_queue.put("server")
                    return
    except Exception as e:
        error_queue.put([e, format_exc()])

def start_server_replay(map_path, replay_path, recv_queue, visual_queue, client_queues):
    ReplayManager.set_input(replay_path)
    NetworkManager.set_queues(visual_queue, recv_queue, client_queues)
    
    # Send initial map data.
    NetworkManager.send_internal_obj(eval(ReplayManager.read_line()), replay=False)
    # Send player data
    NetworkManager.send_internal_obj(eval(ReplayManager.read_line()), replay=False)

    # Sample mainloop
    while True:
        line = eval(ReplayManager.read_line())
        NetworkManager.send_internal_obj(line, replay=False)
        NetworkManager.wait_visual_response()
        if line["type"] == "winner":
            return



import numpy as np
import random
from numpy.linalg import norm
from abc import ABC
from enum import Enum, auto

from codequest22.server.global_map import GlobalMap
from codequest22.stats.ants import Worker, Fighter, Settler

class Ant(ABC):

    def __init__(self, player_index, id, position, energy, color) -> None:
        stats = AntTypes.get_stats(self.TYPE)
        self.color = color
        self.player_index = player_index
        self.id = id
        self.position = position
        self.hp = stats.HP
        self.speed = stats.SPEED
        self.cost = energy
        if hasattr(stats, "LIFESPAN"):
            self.ticks_left = stats.LIFESPAN
        self.goal = None
        self.moved = False
        self.defined_keys = [
            "color",
            "player_index",
            "id",
            "position",
            "hp",
            "cost",
        ]

    def tick(self) -> None:
        self.moved = False
        self.old_position = self.position
        if hasattr(self, "ticks_left"):
            self.ticks_left -= 1
        if not self.alive(): return
        if self.goal is not None:
            togo = self.calculate_speed()
            while self._follow_path:
                new_pos = np.array(list(map(float, self._follow_path[-1])))
                if len(self._follow_path) > 1:
                    new_pos += 0.8*np.array([0.5-random.random(), 0.5-random.random()])
                move_vec = new_pos - np.array(self.position)
                normal = norm(move_vec)
                if normal > togo:
                    self.position = np.array(self.position) + move_vec * (togo / normal)
                    break
                self.position = new_pos
                togo -= normal
                self._follow_path.pop()
                if togo <= 0:
                    break
            # I could be in a bad position atm.
            ipos = (round(self.position[0]), round(self.position[1]))
            if GlobalMap.map_obj[ipos[0]][ipos[1]] == "W":
                # ISSUE: We need to backtrack in some direction from our move_vec.
                options = []
                if move_vec[0] > 0:
                    options.append((1, 0))
                elif move_vec[0] < 0:
                    options.append((-1, 0))
                if move_vec[1] > 0:
                    options.append((0, 1))
                elif move_vec[1] < 0:
                    options.append((0, -1))
                best = None
                for dx, dy in options:
                    if GlobalMap.map_obj[ipos[0]+dx][ipos[1]+dy] != "W":
                        # We can go here.
                        if dx != 0:
                            fixed_pos = (int(self.position[0]+dx) - 0.5 * dx, self.position[1])
                            distance = pow(fixed_pos[0] - new_pos[0], 2) + pow(fixed_pos[1] - new_pos[1], 2)
                            if best is None or best[1] > distance:
                                best = (fixed_pos, distance)
                        if dy != 0:
                            fixed_pos = (self.position[0], int(self.position[1]+dy) - 0.5 * dy)
                            distance = pow(fixed_pos[0] - new_pos[0], 2) + pow(fixed_pos[1] - new_pos[1], 2)
                            if best is None or best[1] > distance:
                                best = (fixed_pos, distance)
                if best is None:
                    print(ipos)
                    print("ERROR: A pathfinding error occured. An ant was placed on a non-traversable tile and couldn't be shifted. Just going to move it to the next follow point.")
                    self.position = self._follow_path[-1] if self._follow_path else self.goal
                else:
                    self.position = fixed_pos
            if not self._follow_path:
                self.goal = None
            self.moved = True

    def alive(self) -> bool:
        return self.hp > 0 and self.ticks_left > 0

    def set_goal(self, g):
        self.goal = (g[0], g[1])
        pos = (self.position[0], self.position[1])
        self._follow_path = GlobalMap.get_path(pos, self.goal)[::-1]

    def calculate_speed(self):
        return self.speed

    def __repr__(self):
        return f"{ {key: getattr(self, key) for key in self.defined_keys} }"

    def to_json(self):
        self.position = tuple(self.position)
        return {
            "classname": self.__class__.__name__,
            "info": {key: getattr(self, key) for key in self.defined_keys}
        }
    
    @classmethod
    def from_json(cls, data):
        ant = cls(data["info"]["player_index"], data["info"]["id"], data["info"]["position"], data["info"]["cost"], data["info"]["color"])
        for key in data["info"]:
            setattr(ant, key, data["info"][key])
        return ant

class AntTypes(Enum):
    WORKER = auto()
    FIGHTER = auto()
    SETTLER = auto()

    def get_class(type: "AntTypes") -> Ant:
        if type == AntTypes.WORKER:
            return WorkerAnt
        elif type == AntTypes.FIGHTER:
            return FighterAnt
        elif type == AntTypes.SETTLER:
            return SettlerAnt
        raise ValueError(f"Unknown ant type {type}")

    def get_stats(type: "AntTypes"):
        if type == AntTypes.WORKER or str(type).strip("'") == "WORKER":
            return Worker
        elif type == AntTypes.FIGHTER or str(type).strip("'") == "FIGHTER":
            return Fighter
        elif type == AntTypes.SETTLER or str(type).strip("'") == "SETTLER":
            return Settler
        raise ValueError(f"Unknown ant type {type}")

    def __repr__(self) -> str:
        return f"\'{self.name}\'"

    def to_json(self):
        return self.__repr__()

class WorkerAnt(Ant):
    TYPE = AntTypes.WORKER

    def __init__(self, player_index, id, position, energy, color) -> None:
        super().__init__(player_index, id, position, energy, color)
        self.encumbered = False
        self.encumbered_energy = 0
        stats = AntTypes.get_stats(AntTypes.WORKER)
        self.remaining_trips = stats.TRIPS
        self.work_rate = stats.WORK_RATE
        self.encumbered_rate = stats.ENCUMBERED_RATE
        self.defined_keys.append("encumbered_energy")

    def alive(self) -> bool:
        return self.hp > 0 and self.remaining_trips > 0

    def free_energy(self):
        self.encumbered = False
        self.encumbered_energy = 0
        self.remaining_trips -= 1
    
    def calculate_speed(self):
        s = super().calculate_speed()
        if self.encumbered:
            return s * self.encumbered_rate
        return s

class FighterAnt(Ant):
    TYPE = AntTypes.FIGHTER

    def __init__(self, player_index, id, position, energy, color) -> None:
        super().__init__(player_index, id, position, energy, color)
        stats = AntTypes.get_stats(AntTypes.FIGHTER)
        self.attack_damage = stats.ATTACK
        self.range = stats.RANGE
        self.num_attacks = stats.NUM_ATTACKS
        self.defined_keys.append("ticks_left")

    def attack(self, other_ant) -> None:
        other_ant.hp -= self.attack_damage

class SettlerAnt(Ant):
    TYPE = AntTypes.SETTLER

    def __init__(self, player_index, id, position, energy, color) -> None:
        super().__init__(player_index, id, position, energy, color)
        self.defined_keys.append("ticks_left")



import json
from queue import Empty
from typing import Optional
from multiprocessing import Queue

RECV_QUEUE: Optional[Queue] = None
SEND_QUEUE: Optional[Queue] = None

IS_REPLAY = None
REPLAY_PATH = None
REPLAY_LINES = None
CUR_REPLAY_INDEX = None

# Import JSON Serializable custom classes.
from codequest22.server.energy import EnergyTile
from codequest22.server.events import *
from codequest22.server.requests import *
from codequest22.server.ant import *

def _rec_finish_dict(object):
    # Recursively fill in json data with appropriate classes
    if isinstance(object, list):
        for i, o in enumerate(object):
            object[i] = _rec_finish_dict(o)
    elif isinstance(object, dict):
        for k, o in object.items():
            object[k] = _rec_finish_dict(o)
        if "classname" in object:
            return eval(object["classname"]).from_json(object)
    return object

def use_queue(recv, send):
    global RECV_QUEUE, SEND_QUEUE, IS_REPLAY
    IS_REPLAY = False
    RECV_QUEUE = recv
    SEND_QUEUE = send

def use_replay(replay_path):
    global IS_REPLAY, REPLAY_PATH, REPLAY_LINES, CUR_REPLAY_INDEX
    IS_REPLAY = True
    REPLAY_PATH = replay_path
    with open(replay_path, "r") as f:
        REPLAY_LINES = f.readlines()
    CUR_REPLAY_INDEX = 0

def is_replay():
    return IS_REPLAY

def send(data):
    SEND_QUEUE.put(data)

def get_input_wait():
    global CUR_REPLAY_INDEX
    if IS_REPLAY:
        l = json.loads(REPLAY_LINES[CUR_REPLAY_INDEX])
        l = _rec_finish_dict(l)
        CUR_REPLAY_INDEX += 1
        return l
    return RECV_QUEUE.get()

def get_input_maybe():
    global CUR_REPLAY_INDEX
    if IS_REPLAY and CUR_REPLAY_INDEX < len(REPLAY_LINES):
        l = json.loads(REPLAY_LINES[CUR_REPLAY_INDEX])
        l = _rec_finish_dict(l)
        CUR_REPLAY_INDEX += 1
        return l
    elif IS_REPLAY:
        return None
    try:
        return RECV_QUEUE.get_nowait()
    except Empty:
        return None



from abc import ABC
from codequest22.server.ant import Ant, WorkerAnt
import codequest22.stats as stats

class Event(ABC):
    
    def __repr__(self) -> str:
        return f"{{\"{self.__class__.__name__}\": {self.get_args()}}}"
    
    def to_json(self):
        return {
            "classname": self.__class__.__name__,
            "args": self.get_args(),
        }
    
    @classmethod
    def from_json(cls, data):
        return cls(*data["args"])

class SpawnEvent(Event):
    """
    An event that means an ant has been spawned.
    Attributes:
    * player_index: The index of the player this ant was spawned from (0-3)
    * ant_id: The particular id assigned to this ant
    * ant_type: The type of ant spawned. Is one of AntTypes
    * position: The spawn position of the ant
    * hp: The starting hp of the ant
    * cost: The energy cost incurred by the player
    """
    def __init__(self, ant: Ant, cost: int, color=None, path=None, goal=None) -> None:
        self.ant_str = ant.to_json()
        self.player_index = ant.player_index
        self.ant_id = ant.id
        self.ant_type = ant.TYPE
        self.position = ant.position[::-1]
        self.hp = ant.hp
        self.cost = cost
        self.color = color
        self.path = path
        self.goal = goal
        if hasattr(ant, "ticks_left"):
            self.ticks_left = ant.ticks_left
        if hasattr(ant, "remaining_trips"):
            self.remaining_trips = ant.remaining_trips

    def get_args(self):
        return [self.ant_str, self.cost, self.color, self.path, self.goal]

class MoveEvent(Event):
    """
    An event that means an ant has moved in the last tick.
    Attributes:
    * player_index: The index of the player this ant was spawned from (0-3)
    * ant_id: The particular id assigned to this ant
    * position: The current position of the ant
    """
    def __init__(self, ant: Ant, path=None) -> None:
        self.ant_str = ant.to_json()
        self.player_index = ant.player_index
        self.ant_id = ant.id
        self.position = ant.position[::-1]
        self.path = path

    def get_args(self):
        return [self.ant_str, self.path]

class DieEvent(Event):
    """
    An event that means an ant has died in the last tick.
    Attributes:
    * player_index: The index of the player this ant was spawned from (0-3)
    * ant_id: The particular id assigned to this ant
    * old_age: True if the ant died of end of lifespan, and False if killed by losing hp
    """
    def __init__(self, ant: Ant) -> None:
        self.ant_str = ant.to_json()
        self.player_index = ant.player_index
        self.ant_id = ant.id
        self.old_age = ant.hp > 0
    
    def get_args(self):
        return [self.ant_str]

class AttackEvent(Event):
    """
    An event that means an ant has been attacked by another ant.
    Attributes:
    * attacker_index: The index of the player the attacker ant was spawned from (0-3)
    * defender_index: The index of the player the defender ant was spawned from (0-3)
    * attacker_id: The particular id assigned to the attacker ant
    * defender_id: The particular id assigned to the defender ant
    * defender_hp: The new hp of the defending ant
    """
    def __init__(self, ant1: Ant, ant2: Ant) -> None:
        self.ant1_str = ant1.to_json()
        self.ant2_str = ant2.to_json()
        self.attacker_index = ant1.player_index
        self.defender_index = ant2.player_index
        self.attacker_id = ant1.id
        self.defender_id = ant2.id
        self.defender_hp = ant2.hp

    def get_args(self):
        return [self.ant1_str, self.ant2_str]

class DepositEvent(Event):
    """
    An event that means a worker ant has deposited it's held energy to the queen.
    Attributes:
    * player_index: The index of the player this ant was spawned from (0-3)
    * ant_id: The particular id assigned to this ant
    * cur_energy: The old energy level of the player
    * energy_amount: The amount of energy deposited by the ant
    * total_energy: The new energy level of the player
    """
    def __init__(self, ant: WorkerAnt, cur_energy: int) -> None:
        self.ant_str = ant.to_json()
        self.cur_energy = cur_energy
        self.player_index = ant.player_index
        self.ant_id = ant.id
        self.energy_amount = int(ant.encumbered_energy)
        self.total_energy = min(cur_energy + self.energy_amount, stats.general.MAX_ENERGY_STORED)

    def get_args(self):
        return [self.ant_str, self.cur_energy]

class ProductionEvent(Event):
    """
    An event that means an ant has been given energy by the food tile.
    Attributes:
    * player_index: The index of the player this ant was spawned from (0-3)
    * ant_id: The particular id assigned to this ant
    * energy_amount: The amount of energy the ant has received
    """
    def __init__(self, ant: WorkerAnt) -> None:
        self.ant_str = ant.to_json()
        self.player_index = ant.player_index
        self.ant_id = ant.id
        self.energy_amount = ant.encumbered_energy
    
    def get_args(self):
        return [self.ant_str]

class ZoneActiveEvent(Event):
    """
    An event that means a settler zone is now active.
    Attributes:
    * points: A list of all points in this settler zone
    * num_ticks: The number of ticks this zone will remain active
    """
    def __init__(self, zone_index: int, num_ticks: int, points: list) -> None:
        self.zone_index = zone_index
        self.points = [tuple(p) for p in points]
        self.num_ticks = num_ticks

    def get_args(self):
        return [self.zone_index, self.num_ticks, self.points]

class ZoneDeactivateEvent(Event):
    """
    An event that means a settler zone is no longer active.
    Attributes:
    * points: A list of all points in this settler zone
    """
    def __init__(self, zone_index: int, points: list) -> None:
        self.zone_index = zone_index
        self.points = [tuple(p) for p in points]

    def get_args(self):
        return [self.zone_index, self.points]

class FoodTileActiveEvent(Event):
    """
    An event that means a food tile is now supercharged.
    Attributes:
    * pos: The position of the food tile
    * num_ticks: The number of ticks the food tile will remain supercharged
    * multiplier: The multiplier applied to all energy received from this tile. Currently always 2.
    """
    def __init__(self, pos: tuple, num_ticks: int, multiplier: int) -> None:
        self.pos = tuple(pos)
        self.num_ticks = num_ticks
        self.multiplier = multiplier

    def get_args(self):
        return [self.pos, self.num_ticks, self.multiplier]

class FoodTileDeactivateEvent(Event):
    """
    An event that means a food tile is no longer supercharged
    Attributes:
    * pos: The position of the food tile
    """
    def __init__(self, pos: tuple) -> None:
        self.pos = pos

    def get_args(self):
        return [self.pos]

class SettlerScoreEvent(Event):
    """
    An event that means a settler ant has received hill score for sitting on an active hill zone.
    Attributes:
    * player_index: The index of the player this ant was spawned from (0-3)
    * ant_id: The particular id assigned to this ant
    * score_amount: The amount of score the player has received for this action
    """
    def __init__(self, ant: Ant, score: int) -> None:
        self.ant_str = ant.to_json()
        self.player_index = ant.player_index
        self.ant_id = ant.id
        self.score_amount = score

    def get_args(self):
        return [self.ant_str, self.score_amount]

class QueenAttackEvent(Event):
    """
    An event that means a fighter ant has successfully attacked the queen
    Attributes:
    * ant_player_index: The index of the player whose fighter ant is attacking (0-3)
    * queen_player_index: The index of the player whose queen is being attacked (0-3)
    * ant_id: The particular id assigned to the fighter ant
    * queen_hp: The new queen hp
    """
    def __init__(self, ant: Ant, queen_index: int, queen_hp: int):
        self.ant_str = ant.to_json()
        self.ant_player_index = ant.player_index
        self.ant_id = ant.id
        self.queen_player_index = queen_index
        self.queen_hp = queen_hp

    def get_args(self):
        return [self.ant_str, self.queen_player_index, self.queen_hp]

class TeamDefeatedEvent(Event):
    """
    An event that means a particular queen has lost all of its health and the team are defeated.
    Attributes:
    * defeated_index: The index of the player who was defeated (0-3)
    * by_index: The index of the player who defeated them (0-3)
    * new_hill_score: The new hill score of the player who defeated them
    """
    def __init__(self, defeated_index: int, by_index: int, hill_score: int) -> None:
        self.defeated_index = defeated_index
        self.by_index = by_index
        self.new_hill_score = hill_score

    def get_args(self):
        return [self.defeated_index, self.by_index, self.new_hill_score]



import os
MAX_ANTS_PER_PLAYER = int(os.environ.get("MAX_ANTS_PER_PLAYER", "100"))
SIMULATION_TICKS = int(os.environ.get("SIMULATION_TICKS", "1200"))
MAX_SPAWNS_PER_TICK = int(os.environ.get("MAX_SPAWNS_PER_TICK", "5"))
MAX_ENERGY_STORED = int(os.environ.get("MAX_ENERGY_STORED", "750"))
QUEEN_HEALTH = int(os.environ.get("QUEEN_HEALTH", "3000"))
STARTING_ENERGY = int(os.environ.get("STARTING_ENERGY", "100"))


from codequest22.client.runner import BotRunner
from codequest22.client.network import NetworkManager
from traceback import format_exc
import os.path

def run_client(bot_path, send_queue, recv_queue, error_queue):
    try:
        NetworkManager.set_queues(send_queue, recv_queue)

        runner = BotRunner()
        runner.load_bot(bot_path)

        team_name = runner.run_command("get_team_name", 0.5)
        folder_path = bot_path
        if folder_path.endswith(".py"):
            folder_path = os.path.dirname(folder_path)
        team_image = os.path.join(folder_path, "profile.png")
        if type(team_name) != str:
            print(f"Team name must be string, got {type(team_name)}.")
            team_name = "Error"
        NetworkManager.send_obj({
            "type": "player_data",
            "name": team_name,
            "image": team_image,
        })

        player_obj = NetworkManager.recv_obj()
        player_index = player_obj["index"]
        runner.run_command("read_index", 0.2, player_index, player_obj["total"])

        map_obj = NetworkManager.recv_obj()
        assert map_obj["type"] == "map"
        energy = map_obj["energy_tiles"]
        map_data = map_obj["obj"]
        runner.run_command("read_map", 30, map_data, energy)

        # Sample mainloop
        while True:
            tick_data = NetworkManager.recv_obj()
            if tick_data["type"] == "finish":
                return
            assert tick_data["type"] == "tick"
            runner.run_command("handle_failed_requests", 0.5, [req for req in tick_data["failed_requests"] if req.player_index == player_index])
            requests = runner.run_command("handle_events", 0.5, tick_data["events"])
            """if not position_valid(position, tick_data["positions"][player_index], map_data):
                print(f"Invalid position {position} received.")
                position = tick_data["positions"][player_index]
                runner.running = False"""
            NetworkManager.send_obj(requests)
    except Exception as e:
        error_queue.put([e, format_exc()])

def position_valid(new_pos, old_pos, map):
    return new_pos == old_pos or new_pos in map["edges"][old_pos]



from abc import ABC
from codequest22.server.ant import AntTypes

class Request(ABC):
    
    def __repr__(self) -> str:
        return f"{{\"{self.__class__.__name__}\": {self.get_args()}}}"
    
    def to_json(self):
        return {
            "classname": self.__class__.__name__,
            "args": self.get_args(),
        }
    
    @classmethod
    def from_json(cls, data):
        return cls(*data["args"])

class SpawnRequest(Request):
    """
    A request to spawn an ant.
    Arguments:
    * ant_type: One of AntTypes.WORKER, AntTypes.SETTLER and AntTypes.FIGHTER
    * id: An id to give the ant. If it is None, then the game will auto allocate one.
    * color: A color for debugging the ant in game (Currently not functional due to some visual bugs)
    * goal: The goal of this ant to move to. Given as tuple of coordinates.
    """
    def __init__(self, ant_type: AntTypes, id=None, color=None, goal=None) -> None:
        super().__init__()
        self.ant_type = ant_type
        self.id = id
        self.color = color
        self.goal = None
        if goal is not None:
            self.goal = goal[1], goal[0]
        self.cost = AntTypes.get_stats(self.ant_type).COST
    
    def get_args(self):
        return [self.ant_type, self.id, self.color, self.goal]

class GoalRequest(Request):
    """
    A request to change the goal of an ant.
    Arguments:
    * id: The id of the ant to move.
    * position: The position the ant should move towards. Given as tuple of coordinates.
    """
    def __init__(self, id, position) -> None:
        super().__init__()
        self.ant_id = id
        self.position = position

    def get_args(self):
        return [self.ant_id, self.position]




import os
import random
import arcade
import math

from arcade.key import MOD_SHIFT
from codequest22.server import energy
from codequest22.server.ant import AntTypes
from codequest22 import stats

import codequest22.visual.constants as constants
from codequest22.visual.utils import resolve_path, hex_to_rgb
from codequest22.visual.game import CleanExit, GameStateHandler

this_dir = os.path.dirname(os.path.abspath(__file__))
f = lambda s: os.path.join(this_dir, s)

TEAM_COLORS = [
    (237, 57, 57),
    (63, 77, 237),
    (248, 218, 84),
    (31, 156, 29),
]

class AntShape:
    def __init__(self, pos, c) -> None:
        self.x, self.y = pos
        self.angle = 0
        self.shape_list = arcade.ShapeElementList()
        self.shape_list.append(arcade.create_rectangle_filled(0, 0, 2, 2, c))

    def draw(self):
        self.shape_list.center_x = self.x
        self.shape_list.center_y = self.y
        self.shape_list.angle = self.angle
        self.shape_list.draw()

class AntCard(arcade.AnimatedTimeBasedSprite):

    def __init__(self, pos, ant_type, team_name, distance=20, scale=0.02):
        super().__init__(scale=scale)
        self.start_pos = [pos[0] - distance, pos[1]]
        self.mid_pos = [pos[0], pos[1]]
        self.current_pos = self.start_pos
        self.forwards = True
        self.angle = 0
        self.center_x, self.center_y = self.start_pos
        tex = arcade.load_texture(resolve_path(f(f"sprites/cards/{team_name}-{ant_type}.jpg")))
        self.frames.append(arcade.AnimationKeyframe(0, 100, tex))

    def update_animation(self, dt=1/60):
        res = super().update_animation(delta_time=dt)
        self.change_y = 0
        self.current_pos = [self.center_x, self.center_y]
        if self.current_pos[0] >= self.mid_pos[0]:
            self.forwards = False
            self.change_x = 0
            self.center_x = self.mid_pos[0]
        if self.forwards:
            self.change_x = max(0.5, pow(self.mid_pos[0] - self.current_pos[0], 1.3)//25)
            return res
        self.alpha = max(self.alpha - 10, 0)
        if self.alpha == 0:
            self.remove_from_sprite_lists()
        return res


class Ant(arcade.AnimatedTimeBasedSprite):

    def __init__(self, pos, c, scale=1):
        super().__init__(scale=scale)
        self.current_pos = pos
        self.new_pos = pos
        self.c_square = AntShape(pos, c)
        self.center_x, self.center_y = pos

    def update(self):
        res = super().update()
        self.c_square.x, self.c_square.y = self.center_x, self.center_y
        return res

    def update_animation(self, delta_time: float = 1 / 60):
        res = super().update_animation(delta_time=delta_time)
        if self.change_x != 0 or self.change_y != 0:
            self.stop_counter -= delta_time
            if self.stop_counter <= 0:
                self.change_x, self.change_y = 0, 0
                self.center_x, self.center_y = self.new_pos
        return res

    def set_new_pos(self, new_pos):
        self.stop_counter = GameStateHandler.get_wait_time()
        self.current_pos = self.new_pos
        self.new_pos = new_pos
        self.change_x = (self.new_pos[0] - self.current_pos[0]) / (60 * GameStateHandler.get_wait_time())
        self.change_y = (self.new_pos[1] - self.current_pos[1]) / (60 * GameStateHandler.get_wait_time())
        self.center_x, self.center_y = self.current_pos
        self.radians = math.atan2(self.change_y, self.change_x) - math.pi / 2
        self.c_square.angle = self.radians * 180 / math.pi

class GoalLine(arcade.ShapeElementList):

    def __init__(self, points):
        super().__init__()
        for p1, p2 in zip(points[:-1], points[1:]):
            self.append(arcade.create_line(p1[0], p1[1], p2[0], p2[1], (0, 0, 0), line_width=1))
        for p in points:
            self.append(arcade.create_ellipse_filled(p[0], p[1], 4, 4, (0, 255, 0)))

class ScreenManager(arcade.Window):

    instance : "ScreenManager" = None

    """General Stuff"""
    def __init__(self, width, height, title="CodeQuest", **kwargs):
        ScreenManager.instance = self
        self.vertex_list = None
        self.player_list = None
        self.s_width = width
        self.s_height = height
        super().__init__(width, height, title)
        self.scaling = min(width / 640, height / 480)
        arcade.set_background_color(hex_to_rgb(constants.GREYSCALE[-1]))
        self.setPlayableArea()
        self.ant_list = arcade.SpriteList()
        self.ant_squares = []
        # Debug Info
        self.goal_line = None
        self.goal_following = None
        self.position_info = None
        self.position_info_pane = None
        self.position_info_pane_2 = None
        self.position_info_text = []
        self.emitters = []

        self.deposit_energy_sound = arcade.load_sound(f("sounds/deposit_energy.wav"))
        self.energy_active_sound = arcade.load_sound(f("sounds/energy_active.wav"))
        self.energy_deactive_sound = arcade.load_sound(f("sounds/energy_deactive.wav"))
        self.game_complete_sound = arcade.load_sound(f("sounds/game_complete.wav"))
        self.kill_sound = arcade.load_sound(f("sounds/kill.wav"))
        self.settler_active_sound = arcade.load_sound(f("sounds/settler_active.wav"))
        self.settler_deactive_sound = arcade.load_sound(f("sounds/settler_deactive.wav"))

    def spawnAnt(self, ant):

        pos = self.translate(*ant.position)
        sprite = Ant(pos, ant.color, self.scaling * 0.015)
        sprite.goal = ant.goal
        sprite.path = ant.path
        col_map = ["red", "blue", "yellow", "green"]
        if isinstance(ant, AntTypes.get_class(AntTypes.WORKER)):
            sprite_strings = [
                f"sprites/ant/{col_map[ant.player_index]}-settler.png",
                f"sprites/ant/{col_map[ant.player_index]}-settler2.png",
            ]
            pos = [
                self.player_ui_main_top_left[ant.player_index][0] + 23 * self.scaling, 
                self.player_ui_main_top_left[ant.player_index][1] - self.panel_height*0.8
            ]
            self.ant_list.append(AntCard(pos, "worker", col_map[ant.player_index], distance=60))
        elif isinstance(ant, AntTypes.get_class(AntTypes.FIGHTER)):
            sprite_strings = [
                f"sprites/ant/{col_map[ant.player_index]}-fighter.png",
                f"sprites/ant/{col_map[ant.player_index]}-fighter2.png",
            ]
            pos = [
                self.player_ui_main_top_left[ant.player_index][0] + 23 * self.scaling, 
                self.player_ui_main_top_left[ant.player_index][1] - self.panel_height*0.5
            ]
            self.ant_list.append(AntCard(pos, "fighter", col_map[ant.player_index], distance=60))
        elif isinstance(ant, AntTypes.get_class(AntTypes.SETTLER)):
            sprite_strings = [
                f"sprites/ant/{col_map[ant.player_index]}-settler.png",
                f"sprites/ant/{col_map[ant.player_index]}-settler2.png",
            ]
            pos = [
                self.player_ui_main_top_left[ant.player_index][0] + 23 * self.scaling, 
                self.player_ui_main_top_left[ant.player_index][1] - self.panel_height*0.2
            ]
            self.ant_list.append(AntCard(pos, "settler", col_map[ant.player_index], distance=60))
        for string in sprite_strings:
            tex = arcade.load_texture(resolve_path(string))
            sprite.frames.append(arcade.AnimationKeyframe(0, 200, tex))
        sprite.texture = sprite.frames[0].texture
        self.ant_mapping[ant.player_index][ant.id] = sprite
        self.ant_list.append(sprite)
        self.ant_squares.append(sprite.c_square)

    def moveAnt(self, player_index, key, position, new_path):
        self.ant_mapping[player_index][key].set_new_pos(self.translate(*position))
        self.ant_mapping[player_index][key].path = new_path

    def removeAnt(self, player_index, key):
        self.ant_list.remove(self.ant_mapping[player_index][key])
        self.ant_squares.remove(self.ant_mapping[player_index][key].c_square)
        del self.ant_mapping[player_index][key]

    def setPlayableArea(self):
        self.L_MARG = max(50 * self.scaling, self.s_width/6) + 10 * self.scaling
        self.T_MARG = 30 * self.scaling
        self.R_MARG = 10 * self.scaling
        self.B_MARG = 10 * self.scaling
        self.PLAYABLE_WIDTH = self.s_width - self.L_MARG - self.R_MARG
        self.PLAYABLE_HEIGHT = self.s_height - self.T_MARG - self.B_MARG

    def translate(self, x, y):
        return x * self.SIDELENGTH + self.XOFFSET, -y * self.SIDELENGTH + self.YOFFSET

    def inverse_translate(self, x, y):
        return int((x - self.XOFFSET + self.SIDELENGTH/2) // self.SIDELENGTH), -1 * int((y - self.YOFFSET - self.SIDELENGTH/2) // self.SIDELENGTH)-1

    def initMap(self, map_data):
        self.map_data = map_data
        self.grid_fixes = arcade.ShapeElementList()
        self.grid_bg = arcade.SpriteList(is_static=True)
        self.after_grid = arcade.SpriteList(is_static=True)
        self.active_grid_bg = [arcade.SpriteList(is_static=True) for _ in GameStateHandler.zones]
        self.active_tile = [arcade.SpriteList(is_static=True) for _ in GameStateHandler.energy_info]

        COL_MAP = {
            ".": (16*0, 16*3),
            "R": (16*2, 16*3),
            "B": (16*3, 16*3),
            "G": (16*0, 16*4),
            "Y": (16*4, 16*3),
            "F": (16*1, 16*4),
            "W": (16*0, 16*0),
            "Z": (16*3, 16*6),
        }
        self.SIDELENGTH = min(self.PLAYABLE_WIDTH / len(self.map_data[0]), self.PLAYABLE_HEIGHT / len(self.map_data))
        self.XOFFSET = self.L_MARG + (self.PLAYABLE_WIDTH - self.SIDELENGTH * len(self.map_data[0])) / 2 + self.SIDELENGTH / 2
        self.YOFFSET = self.s_height - self.T_MARG - (self.PLAYABLE_HEIGHT - self.SIDELENGTH * len(self.map_data)) / 2 - self.SIDELENGTH / 2

        zoned = [[False for __ in range(len(self.map_data[0]))] for _ in range(len(self.map_data))]

        for y in range(len(self.map_data)):
            for x in range(len(self.map_data[y])):
                c = self.map_data[y][x]
                a, b = None, None
                before = []
                extra = []
                capture = False
                cx, cy = self.translate(x, y)
                if c == ".":
                    a, b = random.choice([
                        (17*0, 17*0),
                        (17*0, 17*1),
                    ])
                elif c == "R":
                    a, b = random.choice([
                        (17*0, 17*0),
                        (17*0, 17*1),
                    ])
                    self.after_grid.append(arcade.Sprite(
                        f("sprites/red-base.png"),
                        center_x = cx,
                        center_y = cy,
                        scale = self.scaling * 24/512
                    ))
                elif c == "B":
                    a, b = random.choice([
                        (17*0, 17*0),
                        (17*0, 17*1),
                    ])
                    self.after_grid.append(arcade.Sprite(
                        f("sprites/blue-base.png"),
                        center_x = cx,
                        center_y = cy,
                        scale = self.scaling * 24/512
                    ))
                elif c == "Y":
                    a, b = random.choice([
                        (17*0, 17*0),
                        (17*0, 17*1),
                    ])
                    self.after_grid.append(arcade.Sprite(
                        f("sprites/yellow-base.png"),
                        center_x = cx,
                        center_y = cy,
                        scale = self.scaling * 24/512
                    ))
                elif c == "G":
                    a, b = random.choice([
                        (17*0, 17*0),
                        (17*0, 17*1),
                    ])
                    self.after_grid.append(arcade.Sprite(
                        f("sprites/green-base.png"),
                        center_x = cx,
                        center_y = cy,
                        scale = self.scaling * 24/512
                    ))
                elif c == "W":
                    before.append(random.choice([
                        (17*0, 17*0),
                        (17*0, 17*1),
                    ]))
                    # Match some tile.
                    # Match is TL, T, TR, R, BR, B, BL, L
                    water_map = {
                        # 3 Walls
                        "A.AWAWAW": (17*1, 17*0),
                        "AWA.AWAW": (17*2, 17*0),
                        "AWAWAWA.": (17*1, 17*1),
                        "AWAWA.AW": (17*2, 17*1),
                        # Corner bits
                        "AWAW.WAW": (17*1, 17*2),
                        "AWAWAW.W": (17*2, 17*2),
                        "AW.WAWAW": (17*1, 17*3),
                        ".WAWAWAW": (17*2, 17*3),
                        # Path endings
                        "AWA.A.A.": (17*1, 17*4),
                        "A.A.AWA.": (17*2, 17*4),
                        "A.A.A.AW": (17*1, 17*5),
                        "A.AWA.A.": (17*2, 17*5),
                        # Path Turns
                        "A.AW.WA.": (17*3, 17*0),
                        "A.A.AW.W": (17*4, 17*0),
                        "AW.WA.A.": (17*3, 17*1),
                        ".WA.A.AW": (17*4, 17*1),
                        # Paths
                        "AWA.AWA.": (17*5, 17*0),
                        "A.AWA.AW": (17*5, 17*1),
                        # Crossroads
                        ".W.W.W.W": (17*3, 17*5),
                        # By itself
                        "A.A.A.A.": (17*4, 17*5),
                        # Nine patch
                        "A.AWWWA.": (17*3, 17*2),
                        "A.A.AWWW": (17*5, 17*2),
                        "AWWWA.A.": (17*3, 17*4),
                        "WWA.A.AW": (17*5, 17*4),
                        "A.AWWWWW": (17*4, 17*2),
                        "AWWWWWA.": (17*3, 17*3),
                        "WWA.AWWW": (17*5, 17*3),
                        "WWWWA.AW": (17*4, 17*4),
                        "WWWWWWWW": (17*4, 17*3),
                    }
                    points = [(y-1, x-1), (y-1, x), (y-1, x+1), (y, x+1), (y+1, x+1), (y+1, x), (y+1, x-1), (y, x-1)]
                    for key in water_map:
                        bad = False
                        for i2, (z, w) in enumerate(points):
                            if not (0 <= z < len(self.map_data) and 0 <= w < len(self.map_data[0])):
                                if key[i2] in "WA": continue
                                bad = True
                                break
                            c2 = self.map_data[z][w]
                            if key[i2] == "W":
                                if c2 in ".RBYGPZ":
                                    bad = True
                                    break
                            if key[i2] == ".":
                                if c2 in "W":
                                    bad = True
                                    break
                        if not bad:
                            a, b = water_map[key]
                    if a == None:
                        # Fallback, show errors.
                        a, b = (9*17, 3*17)
                elif c == "F":
                    a, b = random.choice([
                        (17*0, 17*0),
                        (17*0, 17*1),
                    ])
                    self.after_grid.append(arcade.Sprite(
                        f("sprites/normal.png"),
                        center_x = cx,
                        center_y = cy,
                        scale = self.scaling * 16/512
                    ))
                    for i2, t in enumerate(GameStateHandler.energy_info):
                        if t.position == (y, x):
                            self.active_tile[i2].append(arcade.Sprite(
                                f("sprites/super.png"),
                                center_x = cx,
                                center_y = cy,
                                scale = self.scaling * 16/512
                            ))
                            break
                    else:
                        raise ValueError("Map loading failed...")
                    found = False
                    for z, w in [(y+1, x), (y-1, x), (y, x+1), (y, x-1)]:
                        if 0 <= z < len(self.map_data) and 0 <= w < len(self.map_data[0]):
                            if self.map_data[z][w] == "Z":
                                found = True
                    zoned[y][x] = found
                    if not found:
                        a, b = random.choice([
                            (17*0, 17*0),
                            (17*0, 17*1),
                        ])
                if a == None:
                    capture = True
                    # We are a capture zone
                    before.append(random.choice([
                        (17*0, 17*0),
                        (17*0, 17*1),
                    ]))
                    # Match some tile.
                    # Match is TL, T, TR, R, BR, B, BL, L
                    zone_map = {
                        # 3 Walls
                        "A.AWAWAW": (17*6, 17*0),
                        "AWA.AWAW": (17*7, 17*0),
                        "AWAWAWA.": (17*6, 17*1),
                        "AWAWA.AW": (17*7, 17*1),
                        # Corner bits
                        "AWAW.WAW": (17*6, 17*2),
                        "AWAWAW.W": (17*7, 17*2),
                        "AW.WAWAW": (17*6, 17*3),
                        ".WAWAWAW": (17*7, 17*3),
                        # Path endings
                        "AWA.A.A.": (17*6, 17*4),
                        "A.A.AWA.": (17*7, 17*4),
                        "A.A.A.AW": (17*6, 17*5),
                        "A.AWA.A.": (17*7, 17*5),
                        # Path Turns
                        "A.AW.WA.": (17*8, 17*0),
                        "A.A.AW.W": (17*9, 17*0),
                        "AW.WA.A.": (17*8, 17*1),
                        ".WA.A.AW": (17*9, 17*1),
                        # Paths
                        "AWA.AWA.": (17*9, 17*0),
                        "A.AWA.AW": (17*9, 17*1),
                        # Crossroads
                        ".W.W.W.W": (17*8, 17*5),
                        # By itself
                        "A.A.A.A.": (17*9, 17*5),
                        # Nine patch
                        "A.AWWWA.": (17*8, 17*2),
                        "A.A.AWWW": (17*10, 17*2),
                        "AWWWA.A.": (17*8, 17*4),
                        "WWA.A.AW": (17*10, 17*4),
                        "A.AWWWWW": (17*9, 17*2),
                        "AWWWWWA.": (17*8, 17*3),
                        "WWA.AWWW": (17*10, 17*3),
                        "WWWWA.AW": (17*9, 17*4),
                        "WWWWWWWW": (17*9, 17*3),
                    }
                    points = [(y-1, x-1), (y-1, x), (y-1, x+1), (y, x+1), (y+1, x+1), (y+1, x), (y+1, x-1), (y, x-1)]
                    for key in zone_map:
                        bad = False
                        for i2, (z, w) in enumerate(points):
                            if not (0 <= z < len(self.map_data) and 0 <= w < len(self.map_data[0])):
                                if key[i2] in "WA": continue
                                bad = True
                                break
                            c2 = self.map_data[z][w]
                            if key[i2] == "W":
                                if c2 in ".RBYGW":
                                    bad = True
                                    break
                            if key[i2] == ".":
                                if c2 in "ZF":
                                    bad = True
                                    break
                        if not bad:
                            a, b = zone_map[key]
                    if a == None:
                        # Fallback, show errors.
                        a, b = (4*17, 3*17)
                for p, l in before:
                    s = arcade.Sprite(
                        f("sprites/full_tileset.png"), 
                        image_x=p, 
                        image_y=l, 
                        image_width=16, 
                        image_height=16, 
                        center_x=cx, 
                        center_y=cy,
                        scale=self.SIDELENGTH / 16,
                    )
                    self.grid_bg.append(s)
                s = arcade.Sprite(
                    f("sprites/full_tileset.png"), 
                    image_x=a, 
                    image_y=b, 
                    image_width=16, 
                    image_height=16, 
                    center_x=cx, 
                    center_y=cy,
                    scale=self.SIDELENGTH / 16,
                )
                self.grid_bg.append(s)
                if capture:
                    for i in range(len(GameStateHandler.zones)):
                        if (y, x) in GameStateHandler.zones[i]:
                            s2 = arcade.Sprite(
                                f("sprites/full_tileset.png"), 
                                image_x=a, 
                                image_y=b + 6*17, 
                                image_width=16, 
                                image_height=16, 
                                center_x=cx, 
                                center_y=cy,
                                scale=self.SIDELENGTH / 16,
                            )
                            self.active_grid_bg[i].append(s2)
                for p, l in extra:
                    s = arcade.Sprite(
                        f("sprites/full_tileset.png"), 
                        image_x=p, 
                        image_y=l, 
                        image_width=16, 
                        image_height=16, 
                        center_x=cx, 
                        center_y=cy,
                        scale=self.SIDELENGTH / 16,
                    )
                    self.after_grid.append(s)
        for y in range(1, len(self.map_data)):
            for x in range(1, len(self.map_data[y])):
                # Am I the bottom right of a square of walls/zones? If so then add a little square to hide the bad stuff.
                if self.map_data[y][x] == "W":
                    for a, b in [(y-1, x), (y-1, x-1), (y, x-1)]:
                        if self.map_data[a][b] != "W":
                            break
                    else:
                        px, py = self.translate(x, y)
                        self.grid_fixes.append(arcade.create_rectangle_filled(px-self.SIDELENGTH/2, py+self.SIDELENGTH/2, self.SIDELENGTH*.45, self.SIDELENGTH*.45, (168, 182, 183)))
                if self.map_data[y][x] == "Z" or (self.map_data[y][x] == "F" and zoned[y][x]):
                    for a, b in [(y-1, x), (y-1, x-1), (y, x-1)]:
                        if not (
                            self.map_data[a][b] == "Z" or
                            (self.map_data[a][b] == "F" and zoned[a][b])
                        ):
                            break
                    else:
                        px, py = self.translate(x, y)
                        self.grid_fixes.append(arcade.create_rectangle_filled(px-self.SIDELENGTH/2, py+self.SIDELENGTH/2, self.SIDELENGTH*.45, self.SIDELENGTH*.45, (230, 218, 191)))

    def initPlayers(self, images):
        self.setup_player_ui_main(len(images))
        self.player_list = arcade.SpriteList()
        self.ui_player_list = arcade.SpriteList(is_static=True)
        self.ui_player_list_hill_icon = arcade.SpriteList(is_static=True)
        self.ui_player_list_energy_icon = arcade.SpriteList(is_static=True)
        self.ant_mapping = [{} for _ in range(len(images))]
        names = ["red", "blue", "yellow", "green"]
        for x, image in enumerate(images):
            spritepath = resolve_path(image, os.path.join(os.path.dirname(os.path.dirname(__file__)), "codequest.png"))
            # UI
            ui_sprite = arcade.Sprite(spritepath, scale=self.scaling)
            ratio = ui_sprite.width / ui_sprite.height
            max_height = min(self.player_sprite_height, self.player_sprite_width / ratio)
            ui_sprite.height = max_height
            ui_sprite.width = ui_sprite.height * ratio
            ui_sprite.center_x = self.player_ui_main_top_left[x][0] + self.player_ui_main_width * 0.75
            ui_sprite.center_y = self.player_ui_main_top_left[x][1] - self.player_sprite_height / 2
            hill = arcade.Sprite(f(f"sprites/{names[x]}-base.png"), 
                center_x=self.player_ui_main_top_left[x][0] + self.player_ui_main_width * 0.5, 
                center_y=self.player_ui_main_top_left[x][1] - self.panel_height * 0.64,
                scale=self.scaling * 16/512,
            )
            energy = arcade.Sprite(f("sprites/full_tileset.png"), 
                image_x=0, 
                image_y=17*6, 
                image_width=16, 
                image_height=16, 
                center_x=self.player_ui_main_top_left[x][0] + self.player_ui_main_width * 0.5, 
                center_y=self.player_ui_main_top_left[x][1] - self.panel_height * 0.84,
                scale=self.scaling,
            )
            self.ui_player_list.append(ui_sprite)
            self.ui_player_list_hill_icon.append(hill)
            self.ui_player_list_energy_icon.append(energy)

    def collectEnergyAnimation(self, player_index, ant_id):
        position = self.ant_mapping[player_index][ant_id].new_pos
        self.emitters.append(arcade.Emitter(
            center_xy=position,
            emit_controller=arcade.EmitBurst(3),
            particle_factory=lambda emitter: arcade.FadeParticle(
                filename_or_texture=f("sprites/energy_particle.png"),
                change_xy=arcade.rand_in_circle((0.0, 0.0), 1),
                lifetime=random.uniform(0.3, 0.6),
                scale=2,
                start_alpha=80,
                end_alpha=0
            )
        ))

    def killAnimation(self, player_index, ant_id, old_age):
        position = self.ant_mapping[player_index][ant_id].new_pos
        self.emitters.append(arcade.Emitter(
            center_xy=position,
            emit_controller=arcade.EmitBurst(3),
            particle_factory=lambda emitter: arcade.FadeParticle(
                filename_or_texture=f("sprites/old_age_particle.png") if old_age else f("sprites/dead_particle.png"),
                change_xy=arcade.rand_in_circle((0.0, 0.0), 1),
                lifetime=random.uniform(0.3, 0.6),
                scale=2,
                start_alpha=80,
                end_alpha=0
            )
        ))
        if not old_age:
            arcade.play_sound(self.kill_sound, 0.1, 0, False)

    def playDepositSound(self):
        arcade.play_sound(self.deposit_energy_sound, 0.01, 0, False)

    def playEnergyActiveSound(self):
        arcade.play_sound(self.energy_active_sound, 0.02, 0, False)
    
    def playEnergyDeactiveSound(self):
        arcade.play_sound(self.energy_deactive_sound, 0.02, 0, False)
    
    def playZoneActiveSound(self):
        arcade.play_sound(self.settler_active_sound, 0.1, 0, False)
    
    def playZoneDeactiveSound(self):
        arcade.play_sound(self.settler_deactive_sound, 0.1, 0, False)
    
    def playGameCompleteSound(self):
        arcade.play_sound(self.game_complete_sound, 0.5, 0, False)

    def on_update(self, delta_time: float):
        super().on_update(delta_time)
        if self.ant_list is not None:
            self.ant_list.update()
            for v in self.ant_list:
                v.update_animation(delta_time)
        if self.player_list is not None:
            self.player_list.update()
        if self.goal_following is not None:
            try:
                ant = self.ant_mapping[self.goal_following[0]][self.goal_following[1]]
                self.goal_line = GoalLine(
                    [
                        self.translate(p[1], p[0])
                        for p in ant.path
                    ] + [(ant.center_x, ant.center_y)]
                )
            except Exception as e:
                self.goal_following = None
                self.goal_line = None
        else:
            self.goal_line = None
        to_remove = []
        for i, emitter in enumerate(self.emitters):
            emitter.update()
            if emitter.can_reap():
                to_remove.append(i)
        for idx in to_remove[::-1]:
            del self.emitters[idx]
        if self.position_info is not None:
            px, py = self.translate(*self.position_info)
            t = self.map_data[self.position_info[1]][self.position_info[0]]
            if t == "F":
                tp = tuple(self.position_info[::-1])
                for tile in GameStateHandler.energy_info:
                    if tile.position == tp:
                        text = [
                            f"Food: {tile.amount}"
                        ]
                        break
                else:
                    tile = []
            elif t == "W":
                text = [
                    "Not Traversable"
                ]
            elif t in "RBYG":
                idx = "RBYG".index(t)
                name = ["Red", "Blue", "Yellow", "Green"]
                energy = 0 if idx >= len(GameStateHandler.energy) else GameStateHandler.energy[idx]
                text = [
                    f"{name[idx]} Spawn",
                    f"{energy} Energy"
                ]
            elif t == ".":
                text = [
                    "Traversable"
                ]
            elif t == "Z":
                tp = tuple(self.position_info[::-1])
                for i in range(len(GameStateHandler.zones)):
                    if tp in GameStateHandler.zones[i]:
                        active = GameStateHandler.active_zones[i]
                        break
                else:
                    active = False
                active = "Active" if active else "Not Active"
                text = [
                    f"Hill ({active})"
                ]
            else:
                text = []
            cx, cy = px, py
            if px < self.s_width / 2:
                cx = px + 65 * self.scaling
            else:
                cx = px - 65 * self.scaling
            if py < self.s_height / 2:
                cy = py + 10 * self.scaling * len(text)
            else:
                cy = py - 10 * self.scaling * len(text)
            self.position_info_pane = arcade.create_rectangle_filled(cx, cy, 130 * self.scaling, 20 * self.scaling * len(text), (100, 100, 100, 180))
            self.position_info_pane_2 = arcade.create_rectangle_outline(cx, cy, 130 * self.scaling, 20 * self.scaling * len(text), (150, 150, 150), 1 * self.scaling)
            mid = (len(text)-1) / 2
            self.position_info_text = [
                ((cx - 60 * self.scaling, cy - 5 * self.scaling - 20 * (i - mid) * self.scaling), text[i])
                for i in range(len(text))
            ]
        GameStateHandler.tick(delta_time)

    def on_draw(self):
        arcade.start_render()
        if self.grid_bg is not None:
            self.grid_bg.draw()
        self.draw_player_ui_main_under(GameStateHandler.hill, GameStateHandler.energy, GameStateHandler.health)
        for i in range(len(GameStateHandler.active_zones)):
            if GameStateHandler.active_zones[i]:
                self.active_grid_bg[i].draw()
        if self.grid_fixes is not None:
            self.grid_fixes.draw()
        if self.after_grid is not None:
            self.after_grid.draw()
        for i in range(len(GameStateHandler.active_tiles)):
            if GameStateHandler.active_tiles[i]:
                self.active_tile[i].draw()
        if self.player_list is not None:
            self.player_list.draw()
        if self.ant_list is not None:
            self.ant_list.draw()
        if self.ant_squares is not None:
            for sq in self.ant_squares:
                sq.draw()
        for emitter in self.emitters:
            emitter.draw()
        if self.goal_line is not None:
            self.goal_line.draw()
        if self.position_info is not None:
            if self.position_info_pane is not None:
                self.position_info_pane.draw() 
                self.position_info_pane_2.draw() 
            for (px, py), line in self.position_info_text:
                arcade.draw_text(line, px, py, font_size=10 * self.scaling, font_name=f("fonts/Montserrat/Montserrat-Light.ttf"), color=(255, 255, 255))
        self.draw_player_ui_main(GameStateHandler.hill, GameStateHandler.energy, GameStateHandler.health)

    def on_key_press(self, symbol: int, modifiers: int):
        if symbol == arcade.key.ESCAPE:
            arcade.close_window()
        if symbol == arcade.key.P:
            GameStateHandler.toggleMode()
        if symbol == arcade.key.O:
            GameStateHandler.stepExecution()
        if symbol == arcade.key.F:
            idx = -1 if (modifiers & arcade.key.MOD_SHIFT) else 0
            self.set_fullscreen(not self.fullscreen, arcade.get_screens()[idx])
            self.set_viewport(0, self.s_width, 0, self.s_height)
        if symbol == arcade.key.I:
            GameStateHandler.SPEED_MODE = "SUPER_SLOW" if (modifiers & arcade.key.MOD_SHIFT) else "SLOW"
        if symbol == arcade.key.O:
            GameStateHandler.SPEED_MODE = "SUPER_FAST" if (modifiers & arcade.key.MOD_SHIFT) else "FAST"

    def on_key_release(self, symbol: int, modifiers: int):
        if symbol == arcade.key.I:
            if GameStateHandler.SPEED_MODE in ("SLOW", "SUPER_SLOW"):
                GameStateHandler.SPEED_MODE = "NORMAL"
        if symbol == arcade.key.O:
            if GameStateHandler.SPEED_MODE in ("FAST", "SUPER_FAST"):
                GameStateHandler.SPEED_MODE = "NORMAL"

    def on_mouse_press(self, x: float, y: float, button: int, modifiers: int):
        if self.show_winner:
            # Tabs
            tab_left = self.s_width/2-self.winner_box_total_width/2
            tab_right = self.s_width/2+self.winner_box_total_width/2
            tab_bot = self.s_height/2+self.winner_box_total_height/2
            tab_top = self.s_height/2+self.winner_box_total_height/2+self.tab_height
            if tab_left <= x <= tab_right and tab_bot <= y <= tab_top:
                index = min((x - tab_left) // self.tab_width, len(self.tab_texts))
                self.set_tab(int(index))
            return
        gx, gy = self.inverse_translate(x, y)
        if 0 <= gx < len(self.map_data[0]) and 0 <= gy < len(self.map_data):
            if button == arcade.MOUSE_BUTTON_LEFT:
                # Hide any position info
                self.position_info = None
                # Show goal info
                self.goal_following = None
                best = None
                ant_id = None
                for pid in range(len(self.ant_mapping)):
                    for key in self.ant_mapping[pid]:
                        ant = self.ant_mapping[pid][key]
                        d = abs(x - ant.center_x) + abs(y - ant.center_y)
                        if best is None or d < best:
                            best = d
                            ant_id = (pid, key)
                if best is not None and best < 15 * self.scaling:
                    self.goal_following = ant_id
            elif button == arcade.MOUSE_BUTTON_RIGHT:
                # Hide any goals
                self.goal_following = None
                self.goal_line = None
                # Show position info
                self.position_info = round(gx), round(gy)

    """UI Components"""
    def setup_player_ui_main(self, n_players):
        self.PLAYER_UI_MAIN_Y_MARGIN = 20 * self.scaling
        self.PLAYER_UI_MAIN_X_MARGIN = 10 * self.scaling
        base_height = 160 * self.scaling
        m1 = max(1, n_players - 1)
        self.player_ui_main_width = max(50 * self.scaling, self.s_width/6) - self.PLAYER_UI_MAIN_X_MARGIN
        self.top_ui_size = min(self.s_height * 0.9, base_height * n_players + 20 * m1 * self.scaling)
        self.panel_margins = max(10 * self.scaling, (self.top_ui_size - base_height * n_players) / m1)
        self.panel_height = (self.top_ui_size - self.panel_margins * m1) / n_players
        self.player_ui_main_shapes = arcade.ShapeElementList()
        self.player_ui_main_top_left = [
            (self.PLAYER_UI_MAIN_X_MARGIN, self.s_height - self.PLAYER_UI_MAIN_Y_MARGIN - (self.panel_height + self.panel_margins) * x)
            for x in range(n_players)
        ]
        self.player_sprite_height = self.panel_height * 0.4
        self.player_sprite_width = self.player_ui_main_width * 0.4
        for x in range(n_players):
            self.player_ui_main_shapes.append(arcade.create_rectangle_filled(
                self.player_ui_main_top_left[x][0] + self.player_ui_main_width / 2,
                self.player_ui_main_top_left[x][1] - self.panel_height / 2,
                self.player_ui_main_width,
                self.panel_height,
                (150, 150, 150, 80)
            ))
            self.player_ui_main_shapes.append(arcade.create_line_strip(
                (
                    (self.player_ui_main_top_left[x][0] + self.player_ui_main_width - self.scaling, self.player_ui_main_top_left[x][1]), 
                    (self.player_ui_main_top_left[x][0] + self.player_ui_main_width - self.scaling, self.player_ui_main_top_left[x][1] - self.panel_height)
                ), 
                (255, 255, 255, 255), 
                line_width=2,
            ))
            self.player_ui_main_shapes.append(arcade.create_rectangle_filled(
                self.player_ui_main_top_left[x][0] + self.player_ui_main_width / 2 - self.scaling,
                self.player_ui_main_top_left[x][1] - self.panel_height + self.scaling * 2,
                self.player_ui_main_width - 2 * self.scaling,
                self.scaling * 4,
                (255, 0, 0, 255)
            ))
            self.player_ui_main_shapes.append(arcade.create_rectangle(
                self.player_ui_main_top_left[x][0] + 23 * self.scaling, 
                self.player_ui_main_top_left[x][1] - self.panel_height*0.2,
                16 * self.scaling,
                26 * self.scaling,
                (150, 150, 150, 100),
                border_width=2,
                filled=False,
            ))
            self.player_ui_main_shapes.append(arcade.create_rectangle(
                self.player_ui_main_top_left[x][0] + 23 * self.scaling, 
                self.player_ui_main_top_left[x][1] - self.panel_height*0.5,
                16 * self.scaling,
                26 * self.scaling,
                (150, 150, 150, 100),
                border_width=2,
                filled=False,
            ))
            self.player_ui_main_shapes.append(arcade.create_rectangle(
                self.player_ui_main_top_left[x][0] + 23 * self.scaling, 
                self.player_ui_main_top_left[x][1] - self.panel_height*0.8,
                16 * self.scaling,
                26 * self.scaling,
                (150, 150, 150, 100),
                border_width=2,
                filled=False,
            ))
        self.duration_backing_bar = arcade.create_rectangle_filled(self.L_MARG + self.PLAYABLE_WIDTH / 2, self.s_height - 5 * self.scaling - (self.T_MARG - 10 * self.scaling) /2, self.PLAYABLE_WIDTH * 0.9, (self.T_MARG - 10 * self.scaling), (10, 10, 10))
        self.show_winner = False
        self.winner_tab = 0

    def draw_player_ui_main_under(self, hill, energy, health):
        for x in range(len(energy)):
            arcade.draw_text("W", self.player_ui_main_top_left[x][0] + 15 * self.scaling, self.player_ui_main_top_left[x][1] - self.panel_height*0.8, font_size=11 * self.scaling, font_name=f("fonts/Montserrat/Montserrat-Light.ttf"), align="center", width=16 * self.scaling, anchor_y="center")
            arcade.draw_text("F", self.player_ui_main_top_left[x][0] + 15 * self.scaling, self.player_ui_main_top_left[x][1] - self.panel_height*0.5, font_size=11 * self.scaling, font_name=f("fonts/Montserrat/Montserrat-Light.ttf"), align="center", width=16 * self.scaling, anchor_y="center")
            arcade.draw_text("S", self.player_ui_main_top_left[x][0] + 15 * self.scaling, self.player_ui_main_top_left[x][1] - self.panel_height*0.2, font_size=11 * self.scaling, font_name=f("fonts/Montserrat/Montserrat-Light.ttf"), align="center", width=16 * self.scaling, anchor_y="center")

    def set_tab(self, tab_index):
        self.winner_tab = tab_index
        self.winner_tabs = arcade.ShapeElementList()
        for i in range(len(self.tab_texts)):
            self.winner_tabs.append(arcade.create_rectangle(
                self.s_width/2 + self.tab_width*(1 + 2*i)/2 - self.winner_box_total_width/2,
                self.s_height/2 + 120 * self.scaling + self.tab_height/2,
                self.tab_width,
                self.tab_height,
                (30, 30, 30, 255) if i != tab_index else (90, 90, 90, 255)
            ))

    def draw_player_ui_main(self, hill, energy, health):
        self.player_ui_main_shapes.draw()
        self.ui_player_list.draw()
        for x in range(len(energy)):
            arcade.draw_text(self.names[x][:13], self.player_ui_main_top_left[x][0] + self.player_ui_main_width * 0.3, self.player_ui_main_top_left[x][1] - self.panel_height*0.50, font_size=9 * self.scaling, font_name=f("fonts/Montserrat/Montserrat-Light.ttf"), anchor_y="baseline")
            arcade.draw_text(str(hill[x]), self.player_ui_main_top_left[x][0] + self.player_ui_main_width * 0.65, self.player_ui_main_top_left[x][1] - self.panel_height*0.70, font_size=14 * self.scaling, font_name=f("fonts/Montserrat/Montserrat-Light.ttf"), align="right", width=self.player_ui_main_width*0.3, anchor_y="baseline")
            arcade.draw_text(str(energy[x]), self.player_ui_main_top_left[x][0] + self.player_ui_main_width * 0.65, self.player_ui_main_top_left[x][1] - self.panel_height*0.93, font_size=14 * self.scaling, font_name=f("fonts/Montserrat/Montserrat-Light.ttf"), align="right", width=self.player_ui_main_width*0.3, anchor_y="baseline")
            pct = health[x] / stats.general.QUEEN_HEALTH
            total_width = self.player_ui_main_width - 2 * self.scaling
            arcade.draw_rectangle_filled(
                self.player_ui_main_top_left[x][0] + self.player_ui_main_width / 2 - self.scaling - total_width * (1 - pct) / 2,
                self.player_ui_main_top_left[x][1] - self.panel_height + self.scaling * 2,
                total_width * pct,
                self.scaling * 4,
                (0, 255, 0, 255)
            )
        self.ui_player_list_hill_icon.draw()
        self.ui_player_list_energy_icon.draw()
        self.duration_backing_bar.draw()
        if GameStateHandler.cur_tick == 0:
            arcade.draw_text("Waiting for precomputation", self.L_MARG + self.PLAYABLE_WIDTH * 0.125, self.s_height - 5 * self.scaling - (self.T_MARG - 10 * self.scaling) /2, font_size=14 * self.scaling, font_name=f("fonts/Montserrat/Montserrat-Light.ttf"), align="center", width=self.PLAYABLE_WIDTH * 0.75, anchor_y="center")
        else:
            dist = GameStateHandler.cur_tick / stats.general.SIMULATION_TICKS
            total = sum(GameStateHandler.hill)
            base_line = max(0, min(total//4 -1, min(GameStateHandler.hill) - 0.25 * (max(GameStateHandler.hill) - min(GameStateHandler.hill))))
            vals = GameStateHandler.hill
            if total == 0:
                total = len(GameStateHandler.hill)
                vals = [1]*len(GameStateHandler.hill)
                base_line = 0
            else:
                total -= base_line * len(GameStateHandler.hill)
            base_color = [0, 0, 0, 255]
            for i in range(len(vals)):
                for x in range(3):
                    base_color[x] += TEAM_COLORS[i][x] * ((vals[i] - base_line) / total)
            arcade.draw_rectangle_filled(self.L_MARG + self.PLAYABLE_WIDTH * 0.05 + self.PLAYABLE_WIDTH * 0.45 * dist, self.s_height - 5 * self.scaling - (self.T_MARG - 10 * self.scaling) /2, self.PLAYABLE_WIDTH * 0.9 * dist, (self.T_MARG - 10 * self.scaling), base_color)
            arcade.draw_text(f"{GameStateHandler.cur_tick}/{stats.general.SIMULATION_TICKS}", self.L_MARG + self.PLAYABLE_WIDTH * 0.125, self.s_height - 5 * self.scaling - (self.T_MARG - 10 * self.scaling) /2, font_size=14 * self.scaling, font_name=f("fonts/Montserrat/Montserrat-Light.ttf"), align="center", width=self.PLAYABLE_WIDTH * 0.75, anchor_y="center")
        if self.show_winner:
            self.winner_bg.draw()
            if self.winner_tab == 0:
                arcade.draw_text(
                    self.winner_text, 
                    self.s_width/2,
                    self.s_height/2 - 100 * self.scaling,
                    font_size = self.scaling * 12,
                    anchor_x = "center",
                )
                self.winner_graph_bg.draw()
                for line in self.winner_lines:
                    line.draw()
            elif self.winner_tab == 1:
                name_index = ["Red", "Blue", "Yellow", "Green"]
                arcade.draw_text(
                    "Most aggressive - Fighter ant ratio",
                    self.s_width/2 - self.winner_box_total_width/2 + 10 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 30 * self.scaling,
                    font_size=self.scaling * 12,
                )
                aggr = [(f, i) for i, f in enumerate(GameStateHandler.summary["misc"]["aggressive"]["ratio"])]
                aggr.sort()
                aggr_string = ", ".join(f"{name_index[i]}: {int(f * 100)}%" for f, i in aggr[::-1])
                arcade.draw_text(
                    aggr_string,
                    self.s_width/2 + self.winner_box_total_width/2 - 10 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 50 * self.scaling,
                    font_size=self.scaling * 9,
                    anchor_x="right",
                )
                arcade.draw_text(
                    "Marathon runner - Highest average travelled",
                    self.s_width/2 - self.winner_box_total_width/2 + 10 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 80 * self.scaling,
                    font_size=self.scaling * 12,
                )
                marathon = [(f, i) for i, f in enumerate(GameStateHandler.summary["misc"]["marathon"]["avg"])]
                marathon.sort()
                marathon_string = ", ".join(f"{name_index[i]}: {f:.1f}m" for f, i in marathon[::-1])
                arcade.draw_text(
                    marathon_string,
                    self.s_width/2 + self.winner_box_total_width/2 - 10 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 100 * self.scaling,
                    font_size=self.scaling * 9,
                    anchor_x="right",
                )
                arcade.draw_text(
                    "Holding out - Highest stored energy",
                    self.s_width/2 - self.winner_box_total_width/2 + 10 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 130 * self.scaling,
                    font_size=self.scaling * 12,
                )
                waiting = [(f, i) for i, f in enumerate(GameStateHandler.summary["misc"]["waiting"]["max_energy"])]
                waiting.sort()
                waiting_string = ", ".join(f"{name_index[i]}: {int(f)}E" for f, i in waiting[::-1])
                arcade.draw_text(
                    waiting_string,
                    self.s_width/2 + self.winner_box_total_width/2 - 10 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 150 * self.scaling,
                    font_size=self.scaling * 9,
                    anchor_x="right",
                )
                arcade.draw_text(
                    "Stayin Alive - Highest old age ratio",
                    self.s_width/2 - self.winner_box_total_width/2 + 10 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 180 * self.scaling,
                    font_size=self.scaling * 12,
                )
                alive = [(f, i) for i, f in enumerate(GameStateHandler.summary["misc"]["alive"]["pct"])]
                alive.sort()
                alive_string = ", ".join(f"{name_index[i]}: {100*f:.1f}%" for f, i in alive[::-1])
                arcade.draw_text(
                    alive_string,
                    self.s_width/2 + self.winner_box_total_width/2 - 10 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 200 * self.scaling,
                    font_size=self.scaling * 9,
                    anchor_x="right",
                )
            else:
                arcade.draw_text(
                    "Workers",
                    self.s_width/2 - self.winner_box_total_width/2 + 10 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 30 * self.scaling,
                    font_size=self.scaling * 12,
                )
                arcade.draw_text(
                    f"Spawned: {GameStateHandler.summary['ants']['worker'][self.winner_tab-2]['spawned']}",
                    self.s_width/2 - self.winner_box_total_width/2 + 30 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 50 * self.scaling,
                    font_size=self.scaling * 11,
                )
                arcade.draw_text(
                    f"Deposited: {GameStateHandler.summary['ants']['worker'][self.winner_tab-2]['total_deposited']}E",
                    self.s_width/2 - self.winner_box_total_width/2 + 30 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 65 * self.scaling,
                    font_size=self.scaling * 11,
                )
                arcade.draw_text(
                    "Fighters",
                    self.s_width/2 - self.winner_box_total_width/2 + 10 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 90 * self.scaling,
                    font_size=self.scaling * 12,
                )
                arcade.draw_text(
                    f"Spawned: {GameStateHandler.summary['ants']['fighter'][self.winner_tab-2]['spawned']}",
                    self.s_width/2 - self.winner_box_total_width/2 + 30 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 110 * self.scaling,
                    font_size=self.scaling * 11,
                )
                arcade.draw_text(
                    f"Kills: {GameStateHandler.summary['ants']['fighter'][self.winner_tab-2]['kills']}",
                    self.s_width/2 - self.winner_box_total_width/2 + 30 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 125 * self.scaling,
                    font_size=self.scaling * 11,
                )
                arcade.draw_text(
                    "Settlers",
                    self.s_width/2 - self.winner_box_total_width/2 + 10 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 150 * self.scaling,
                    font_size=self.scaling * 12,
                )
                arcade.draw_text(
                    f"Spawned: {GameStateHandler.summary['ants']['settler'][self.winner_tab-2]['spawned']}",
                    self.s_width/2 - self.winner_box_total_width/2 + 30 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 170 * self.scaling,
                    font_size=self.scaling * 11,
                )
                arcade.draw_text(
                    f"Points: {GameStateHandler.summary['ants']['settler'][self.winner_tab-2]['points']}H",
                    self.s_width/2 - self.winner_box_total_width/2 + 30 * self.scaling,
                    self.s_height/2 + self.winner_box_total_height/2 - 185 * self.scaling,
                    font_size=self.scaling * 11,
                )
            self.winner_tabs.draw()
            for i, text in enumerate(self.tab_texts):
                arcade.draw_text(
                    text,
                    self.s_width/2 + self.tab_width*(1 + 2*i)/2 - self.winner_box_total_width/2,
                    self.s_height/2 + 120 * self.scaling + self.tab_height/2,
                    color=(255, 255, 255, 255),
                    font_size=self.scaling * 12,
                    anchor_x="center"
                )

    def show_winner_ui(self, indicies):
        self.show_winner = True
        self.winner_box_total_width = 360 * self.scaling
        self.winner_box_total_height = 240 * self.scaling
        self.winner_bg = arcade.create_rectangle(
            self.s_width/2, 
            self.s_height/2, 
            self.winner_box_total_width,
            self.winner_box_total_height,
            (50, 50, 50, 180)
        )
        self.winner_tabs = arcade.ShapeElementList()
        self.tab_width = self.winner_box_total_width / (len(GameStateHandler.energy) + 2)
        self.tab_height = 40 * self.scaling
        name_index = ["Red", "Blue", "Yellow", "Green"]
        names = [name_index[i] for i in indicies]
        self.tab_texts = ["Graph", "Misc"] + name_index[:len(GameStateHandler.energy)]
        self.set_tab(0)

        if len(names) == 0:
            self.winner_text = "Draw!"
        elif len(names) == 1:
            self.winner_text = f"{names[0]} Won!"
        else:
            self.winner_text = f"{' and '.join(names)} Won!"
        # Graph
        self.winner_graph_bg = arcade.create_rectangle(
            self.s_width/2,
            self.s_height/2 + 20 * self.scaling,
            180 * self.scaling,
            160 * self.scaling,
            (50, 50, 50, 230)
        )
        max_hill = max(GameStateHandler.hill)
        # graph is from 0 to max_hill + 10%
        max_hill = max(10, max_hill) * 1.1
        self.winner_lines = [arcade.ShapeElementList() for _ in GameStateHandler.hill]
        for i in range(len(GameStateHandler.hill_snapshots)):
            for j in range(len(GameStateHandler.hill_snapshots[i])):
                cx = self.s_width/2 - 80 * self.scaling + (i / len(GameStateHandler.hill_snapshots)) * 160 * self.scaling
                cy = self.s_height/2 - 50 * self.scaling + (GameStateHandler.hill_snapshots[i][j] / max_hill) * 140 * self.scaling
                self.winner_lines[j].append(arcade.create_ellipse_filled(
                    cx, cy, self.scaling*2, self.scaling*2, TEAM_COLORS[j], num_segments=64,
                ))



import json
import math
from heapq import heappush, heappop

from codequest22.server.energy import EnergyTile

class UnreachableException(Exception):
    pass

class GlobalMap:

    map_obj = None

    @classmethod
    def load_map(cls, map_path):
        with open(map_path, "r") as f:
            obj = json.loads(f.read())
        try:
            cls.player_spawns = [None]*4
            cls.energy_tiles = []
            cls.capture_zones = []
            cls.h = len(obj["map"])
            cls.w = len(obj["map"][0])
            energy_info = obj.get("energy", {})
            new_energy_info = {
                eval(key)[::-1]: value
                for key, value in energy_info.items()
            }
            remap = "XW.GBZRYF"
            cls.map_obj = [[remap[obj["map"][x][y]] for y in range(cls.w)] for x in range(cls.h)]
            for x in range(cls.h):
                for y, c in enumerate(cls.map_obj[x]):
                    for i, c2 in enumerate("RBYG"):
                        if c == c2:
                            cls.player_spawns[i] = (x, y)
                    if c == "F":
                        amount = new_energy_info.get((x, y), -1)
                        cls.energy_tiles.append(EnergyTile((x, y), amount))
            for tile in cls.energy_tiles:
                if tile.amount == -1:
                    print("WARNING: The following production zone has no energy information:", tile.position[::-1])
            # Locate zones with DFS
            dealt = [[False for y in range(cls.w)] for x in range(cls.h)]
            for x in range(cls.h):
                for y in range(cls.w):
                    if cls.map_obj[x][y] == "Z" and not dealt[x][y]:
                        q = [(x, y)]
                        cls.capture_zones.append([])
                        while q:
                            (a, b) = q.pop()
                            cls.capture_zones[-1].append((a, b))
                            dealt[a][b] = True
                            for (c, d) in [(a+1, b),(a-1, b),(a, b+1),(a, b-1)]:
                                if not (0 <= c < cls.h and 0 <= d < cls.w): continue
                                if cls.map_obj[c][d] in "ZF" and not dealt[c][d] and (c, d) not in q:
                                    q.append((c, d))
        except Exception as e:
            print(f"An error occured, while reading the map {map_path}, given below:")
            raise e
        cls._precomp_paths()

    @classmethod
    def get_path(cls, p1, p2):
        if cls.map_obj[round(p1[0])][round(p1[1])] == "W" or cls.map_obj[round(p2[0])][round(p2[1])] == "W":
            raise UnreachableException()
        final_path = []
        to_add = []
        if type(p1[0]) != int or type(p1[1]) != int:
            p1 = (round(p1[0]), round(p1[1]))
            final_path.insert(0, p1)
        if type(p2[0]) != int or type(p2[1]) != int:
            to_add.append(p2)
            p2 = (round(p2[0]), round(p2[1]))
        if p2 not in cls.best_path[p1]:
            raise UnreachableException()
        p = cls.best_path[p1][p2][1]
        while p != p2:
            final_path.append(p)
            p = cls.best_path[p][p2][1]
        final_path.append(p2)
        return final_path + to_add


    @classmethod
    def _precomp_paths(cls):
        # 4*W*H*log(W*H)*W*H = V^2 LOG(V) runtime.
        # So at 30 seconds, Total area = 10^4 (100*100) is a reasonable size to expect students to get everything done.
        cls._gen_special_points()
        # Dijkstra's for every start point.
        # Generate edges
        adj = {}
        cls.best_path = {}
        h, w = len(cls.map_obj), len(cls.map_obj[0])
        points = []
        for y in range(h):
            for x in range(w):
                cls.best_path[(y, x)] = {}
                adj[(y, x)] = []
                if cls.map_obj[y][x] == "W": continue
                points.append((y, x))
        for y, x in points:
            for a, b in [(y+1, x), (y-1, x), (y, x+1), (y, x-1)]:
                if 0 <= a < h and 0 <= b < w and cls.map_obj[a][b] != "W":
                    adj[(y, x)].append((a, b, 1))
            for a, b in [(y+1, x+1), (y+1, x-1), (y-1, x+1), (y-1, x-1)]:
                if (
                    0 <= a < h and 0 <= b < w and 
                    cls.map_obj[a][b] != "W" and
                    cls.map_obj[a][x] != "W" and
                    cls.map_obj[y][b] != "W"
                ):
                    adj[(y, x)].append((a, b, math.sqrt(2)))
        idx = {p: i for i, p in enumerate(points)}
        for y, x in points:
            # Dijkstra run
            expanded = [False] * len(points)
            seen = [False] * len(points)
            queue = []
            heappush(queue, (0, (y, x), (y, x)))
            while queue:
                d, (a, b), parent = heappop(queue)
                if expanded[idx[(a, b)]]: continue
                expanded[idx[(a, b)]] = True
                cls.best_path[(y, x)][(a, b)] = (d, parent)
                # Look at all neighbours
                for j, k, d2 in adj[(a, b)]:
                    if not expanded[idx[(j, k)]]:
                        heappush(queue, (
                            d + d2,
                            (j, k),
                            parent if (parent in cls.special_points and parent != (y, x)) else (j, k)
                        ))

    @classmethod
    def _gen_special_points(cls):
        """
        A special point is one with two traversable tiles adjacent and a non-traversable tile in between.

        This ensures that the shortest path between any tiles on the map can begin with a straight line to a special point.
        """
        cls.special_points = set()
        h, w = len(cls.map_obj), len(cls.map_obj[0])
        for y in range(h):
            for x in range(w):
                points = [(y, x+1), (y+1, x+1), (y+1, x), (y+1, x-1), (y, x-1), (y-1, x-1), (y-1, x), (y-1, x+1), (y, x+1)]
                for z in range(0, 7, 2):
                    if 0 <= points[z][0] < h and 0 <= points[z][1] < w and 0 <= points[z+2][0] < h and 0 <= points[z+2][1] < w:
                        if (
                            cls.map_obj[points[z][0]][points[z][1]] != "W" and
                            cls.map_obj[points[z+1][0]][points[z+1][1]] == "W" and
                            cls.map_obj[points[z+2][0]][points[z+2][1]] != "W"
                        ):
                            cls.special_points.add((y, x))
    




from codequest22.stats import energy

class EnergyTile:

    def __init__(self, position, amount) -> None:
        self.position = position
        self.amount = amount
        self.max_delay = energy.DELAY
        self.per_tick = energy.PER_TICK
        self.cur_delay = 0
    
    def __repr__(self) -> str:
        return f"{{\"{self.__class__.__name__}\": {[self.position, self.amount]}}}"
    
    def to_json(self):
        return {
            "classname": "EnergyTile",
            "position": self.position,
            "amount": self.amount,
        }
    
    @classmethod
    def from_json(cls, data):
        return EnergyTile(tuple(data["position"]), data["amount"])



import arcade
from codequest22.server.ant import AntTypes
from codequest22.server.events import *
from codequest22.server.summarise_game import GameSummariser

from codequest22.visual.game_input import get_input_maybe, send, use_queue, get_input_wait, use_replay, is_replay

class CleanExit(Exception):
    pass

class GameStateHandler:
    
    MODE_PLAY = "play"
    MODE_STOPPED = "stopped"
    CURRENT_MODE = MODE_PLAY
    RESUME_TIME = 1/10
    SPEED_MODE = "NORMAL"
    SCREEN_RES = (1280, 800)

    last_message = 0
    waiting_message = False
    c_time = 0

    @classmethod
    def get_wait_time(cls):
        if cls.SPEED_MODE == "NORMAL":
            return cls.RESUME_TIME
        elif cls.SPEED_MODE == "FAST":
            return 1/3 * cls.RESUME_TIME
        elif cls.SPEED_MODE == "SLOW":
            return 3 * cls.RESUME_TIME
        elif cls.SPEED_MODE == "SUPER_FAST":
            return 1/10 * cls.RESUME_TIME
        elif cls.SPEED_MODE == "SUPER_SLOW":
            return 10 * cls.RESUME_TIME
        raise ValueError(f"SPEED_MODE is invalid: {cls.SPEED_MODE}")

    @classmethod
    def init(cls):
        from codequest22.visual.screen import ScreenManager

        cls.start_payload = get_input_wait()
        assert cls.start_payload["type"] == "map", "Wrong payload encountered in game."
        cls.map_defn = cls.start_payload["obj"]
        cls.zones = [[tuple(y) for y in z] for z in cls.start_payload["zones"]]
        cls.energy_info = cls.start_payload["energy_info"]
        cls.active_zones = [False for _ in cls.zones]
        cls.active_tiles = [False for _ in cls.energy_info]
        cls.sm = ScreenManager(*cls.SCREEN_RES)
        cls.sm.initMap(cls.map_defn)
        cls.cur_tick = 0
        cls.hill_snapshots = []

    @classmethod
    def tick(cls, dt):
        cls.c_time += dt
        if not (is_replay() and cls.c_time - cls.last_message <= cls.get_wait_time()):
            # If we are in a replay, we need to do the waiting ourselves.
            res = get_input_maybe()
            if res is not None:
                if res["type"] == "finish":
                    raise CleanExit()
                elif res["type"] == "player_data":
                    # Get names and images
                    cls.sm.names = res["names"]
                    # Load images as sprites
                    cls.sm.initPlayers(res["images"])
                    cls.energy = res["energy"]
                    cls.hill = res["hill"]
                    cls.health = res["health"]
                elif res["type"] == "tick":
                    cls.cur_tick += 1
                    cls.hill_snapshots.append(cls.hill[::])
                    # Get player information
                    """for i in range(len(cls.sm.names)):
                        cls.sm.movePlayer(i, res["positions"][i])"""
                    for ev in res["events"]:
                        if isinstance(ev, SpawnEvent):
                            ant_obj = AntTypes.get_class(ev.ant_type)(ev.player_index, ev.ant_id, ev.position, ev.cost, ev.color)
                            ant_obj.goal = ev.goal
                            ant_obj.path = ev.path
                            cls.sm.spawnAnt(ant_obj)
                            cls.energy[ev.player_index] -= ev.cost
                        elif isinstance(ev, MoveEvent):
                            cls.sm.moveAnt(ev.player_index, ev.ant_id, ev.position, ev.path)
                        elif isinstance(ev, DieEvent):
                            cls.sm.killAnimation(ev.player_index, ev.ant_id, ev.old_age)
                            cls.sm.removeAnt(ev.player_index, ev.ant_id)
                        elif isinstance(ev, ProductionEvent):
                            cls.sm.collectEnergyAnimation(ev.player_index, ev.ant_id)
                        elif isinstance(ev, DepositEvent):
                            cls.energy[ev.player_index] = ev.total_energy
                            cls.sm.playDepositSound()
                        elif isinstance(ev, ZoneActiveEvent):
                            cls.active_zones[ev.zone_index] = True
                            cls.sm.playZoneActiveSound()
                        elif isinstance(ev, ZoneDeactivateEvent):
                            cls.active_zones[ev.zone_index] = False
                            cls.sm.playZoneDeactiveSound()
                        elif isinstance(ev, FoodTileActiveEvent):
                            for i2, t in enumerate(cls.energy_info):
                                if t.position[::-1] == ev.pos:
                                    cls.active_tiles[i2] = True
                            cls.sm.playEnergyActiveSound()
                        elif isinstance(ev, FoodTileDeactivateEvent):
                            for i2, t in enumerate(cls.energy_info):
                                if t.position[::-1] == ev.pos:
                                    cls.active_tiles[i2] = False
                            cls.sm.playEnergyDeactiveSound()
                        elif isinstance(ev, SettlerScoreEvent):
                            cls.hill[ev.player_index] += ev.score_amount
                        elif isinstance(ev, QueenAttackEvent):
                            cls.health[ev.queen_player_index] = ev.queen_hp
                        elif isinstance(ev, TeamDefeatedEvent):
                            cls.hill[ev.by_index] = ev.new_hill_score
                            cls.hill[ev.defeated_index] = 0
                elif res["type"] == "winner":
                    indicies = res["indicies"]
                    cls.CURRENT_MODE = cls.MODE_STOPPED
                    # Wait for the replay file to be closed, so we can open it.
                    failures = 0
                    while True:
                        try:
                            replay = open(cls.replay_path, "r")
                            replay.close()
                            break
                        except IOError:
                            if failures > 30:
                                raise ValueError("Replay file did not close.")
                            failures += 1
                            import time
                            time.sleep(0.02)
                    cls.summary = GameSummariser.get_results(cls.replay_path)
                    cls.sm.show_winner_ui(indicies)
                    cls.sm.playGameCompleteSound()
                cls.last_message = cls.c_time
                cls.waiting_message = True
        if cls.CURRENT_MODE == cls.MODE_PLAY:
            if cls.waiting_message and cls.c_time - cls.last_message > cls.get_wait_time():
                send("Resume")
                cls.waiting_message = False
        elif cls.CURRENT_MODE == cls.MODE_STOPPED:
            cls.last_message = cls.c_time

    @classmethod
    def toggleMode(cls):
        if cls.CURRENT_MODE == cls.MODE_STOPPED:
            cls.CURRENT_MODE = cls.MODE_PLAY
        else:
            cls.CURRENT_MODE = cls.MODE_STOPPED

    @classmethod
    def stepExecution(cls):
        if cls.CURRENT_MODE == cls.MODE_STOPPED:
            if cls.waiting_message:
                send("Resume")
                cls.waiting_message = False

def run_visual(recv_queue, send_queue, error_queue, is_replay=False, replay_path=""):
    from traceback import format_exc
    try:
        if is_replay:
            use_replay(replay_path)
        else:
            use_queue(recv_queue, send_queue)

        GameStateHandler.init()
        GameStateHandler.replay_path = replay_path
        arcade.run()
        error_queue.put("visual")
    except CleanExit:
        error_queue.put("visual")
    except Exception as e:
        error_queue.put([e, format_exc()])



import math
import os

this_dir = os.path.dirname(os.path.abspath(__file__))
cwd = os.getcwd()

def interpolate_colour_discrete(colours, max_v, min_v, actual):
    proportion = min(1, max(0, (actual - min_v) / (max_v - min_v)))
    index = min(len(colours)-1, math.floor(proportion * len(colours)))
    return colours[index]

def resolve_path(path, fallback=None):
    if path is not None:
        if os.path.isfile(os.path.join(cwd, path)):
            return os.path.join(cwd, path)
        elif os.path.isfile(os.path.join(this_dir, path)):
            return os.path.join(this_dir, path)
    if fallback is None:
        raise ValueError(f"Unknown sprite path {path}")
    return resolve_path(fallback, None)

def hex_to_rgb(hex):
    hex = hex.replace("#", "")
    return tuple(int(hex[i:i+2], 16) for i in range(0, len(hex), 2))



from multiprocessing import Queue

class NetworkManager:

    send: Queue
    recv: Queue

    @classmethod
    def set_queues(cls, send_queue, recv_queue):
        cls.send = send_queue
        cls.recv = recv_queue
    
    @classmethod
    def send_obj(cls, obj):
        cls.send.put(obj)
    
    @classmethod
    def recv_obj(cls):
        return cls.recv.get()



import json
import progressbar
from codequest22.server.replay import ReplayManager
import codequest22.stats as stats

class CustomEncoder(json.JSONEncoder):

    def default(self, o):
        if hasattr(o, "to_json"):
            return o.to_json()
        return super().default(o)

class NetworkManager:

    RECV, SEND = range(2)

    @classmethod
    def set_queues(cls, visual_queue, recv_queue, queues, is_visual):
        cls.visual_queue = visual_queue
        cls.recv_queue = recv_queue
        cls.queues = queues
        cls.is_visual = is_visual
        cls.bar = progressbar.ProgressBar(widgets=[progressbar.Counter(format=f"%(value)d/{stats.general.SIMULATION_TICKS}"), progressbar.Bar()], max_value=stats.general.SIMULATION_TICKS).start()
        cls.tick = 0

    @classmethod
    def send_internal_obj(cls, obj, visual=True, replay=True):
        if visual:
            cls.visual_queue.put(obj)
        if replay:
            ReplayManager.write_line(json.dumps(obj, cls=CustomEncoder))
        if obj["type"] == "tick":
            cls.tick += 1
            cls.bar.update(cls.tick)
        elif obj["type"] == "winner":
            cls.bar.finish()

    @classmethod
    def wait_visual_response(cls):
        if not cls.is_visual: return
        res = cls.recv_queue.get()
        if res == "Resume":
            return
        else:
            raise ValueError("Visual runner exited.")

    @classmethod
    def broadcast_obj(cls, obj):
        for client in cls.queues:
            client[cls.SEND].put(obj)
    
    @classmethod
    def recv_client_response(cls):
        return [ client[cls.RECV].get() for client in cls.queues]

    @classmethod
    def send_client(cls, index, obj):
        cls.queues[index][cls.SEND].put(obj)



class ReplayManager:

    @classmethod
    def set_output(cls, path):
        cls.file = open(path, "w")

    @classmethod
    def set_input(cls, path):
        cls.file = open(path, "r")
        cls.lines = list(cls.file.readlines())
        cls.line_index = 0

    @classmethod
    def write_line(cls, string):
        assert "\n" not in string, "Replay stores should be single lines."
        cls.file.write(string + "\n")

    @classmethod
    def read_line(cls) -> str:
        line = cls.lines[cls.line_index]
        cls.line_index += 1
        return line

    @classmethod
    def close(cls):
        cls.file.close()

class ErrorManager:
    
    @classmethod
    def write_error(cls, path, error_string):
        cls.file = open(path, "w")
        cls.file.write(error_string)



__version__ = "1.1.3"

from codequest22.client import run_client
import argparse, os
from multiprocessing import Process, Queue

parser = argparse.ArgumentParser(description="Run the CodeQuest 2022 simulation.")
parser.add_argument(
    "players",
    type=str,
    nargs="+",
    help="Path to bot scripts for simulating, space separated. If this is 'replay', then it will attempt to play from a replay file instead.",
)
parser.add_argument(
    "--map",
    "-m",
    type=str,
    default="small.json",
    help="The map to simulate on.",
)
parser.add_argument(
    "--no-visual",
    "-nv",
    action="store_false",
    dest="visual",
    help="Whether to open a visual process",
)
parser.add_argument(
    "--replay",
    "-r",
    type=str,
    default="replay.txt",
    help="Path to store or read the replay file.",
)

def main():
    import sys, time
    args = parser.parse_args(sys.argv[1:])
    called_from = os.getcwd()

    if len(args.players) == 1 and args.players[0] == "replay":
        from codequest22.server import start_server_replay as start_server
        is_replay = True
    else:
        from codequest22.server import start_server as start_server
        is_replay = False
        # Resolve bot paths
        full_paths = [None]*len(args.players)
        for i, bot in enumerate(args.players):
            full_paths[i] = os.path.join(called_from, bot)
            if not os.path.isfile(full_paths[i]) and not os.path.isdir(full_paths[i]):
                full_paths[i] = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'client', bot)
            if not os.path.isfile(full_paths[i]) and not os.path.isdir(full_paths[i]):
                raise ValueError(f"Could not find bot file {bot}")

    # Resolve map and replay paths
    map_path = os.path.join(called_from, args.map)
    if not os.path.isfile(map_path):
        map_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'server/maps', args.map)
    if not os.path.isfile(map_path):
        raise ValueError(f"Could not find map file {args.map}")
    replay_path = os.path.join(called_from, args.replay)

    # The server and visual use a queue to communicate.
    # This queue is essentially a replay file from aiarena21, but it allows us to pause/step through execution.
    visual_queue = Queue()
    server_queue = Queue()
    error_queue = Queue()
    client_queues = [] if is_replay else [[Queue(), Queue()] for _ in full_paths]

    server_process = None
    client_processes = []

    def close_everything():
        # Close processes
        for process in [
            server_process,
        ] + client_processes + ([visual_process] if args.visual else []):
            if process is not None:
                process.terminate()
                process.join()
                process.close()
        # Empty queues
        for queue in [
            visual_queue,
            server_queue,
            error_queue,
        ] + [
            q[i]
            for q in client_queues
            for i in range(2)
        ]:
            while not queue.empty():
                queue.get()


    if not is_replay:
        server_process = Process(target=start_server, args=[map_path, replay_path, server_queue, visual_queue, error_queue, client_queues, args.visual])
        client_processes = [
            Process(target=run_client, args=[full_paths[i], *client_queues[i], error_queue])
            for i in range(len(full_paths))
        ]

        visual_process = None
        if args.visual:
            from codequest22.visual import run_visual
            visual_process = Process(target=run_visual, args=[visual_queue, server_queue, error_queue, False, replay_path], daemon=True)

        try:
            server_process.start()
            time.sleep(0.5)
            for p in client_processes:
                p.start()

            if args.visual:
                visual_process.start()

            while True:
                v = error_queue.get()
                if isinstance(v, list) and isinstance(v[0], Exception):
                    from codequest22.server.replay import ErrorManager
                    error_path = replay_path.replace(".txt", ".error_log")
                    ErrorManager.write_error(error_path, v[1])
                    print(f"An error occured. Check {error_path}")
                    close_everything()
                    break                    
                if isinstance(v, str) and v == "visual" and args.visual:
                    close_everything()
                    break
                if isinstance(v, str) and v == "server" and not args.visual:
                    close_everything()
                    break
        except KeyboardInterrupt:
            close_everything()     
        except Exception as e:
            close_everything()
            raise e
    else:
        from codequest22.visual import run_visual
        visual_process = Process(target=run_visual, args=[visual_queue, server_queue, error_queue, True, replay_path], daemon=True)

        try:
            visual_process.start()
            while True:
                v = error_queue.get()
                if isinstance(v, list) and isinstance(v[0], Exception):
                    from codequest22.server.replay import ErrorManager
                    error_path = replay_path.replace(".txt", ".error_log")
                    ErrorManager.write_error(error_path, v[1])
                    print(f"An error occured. Check {error_path}")
                    close_everything()
                    break
                if isinstance(v, str) and v == "visual":
                    close_everything()
                    break
        except KeyboardInterrupt:
            close_everything()     
        except Exception as e:
            from codequest22.server.replay import ErrorManager
            error_path = replay_path.replace(".txt", ".error_log")
            ErrorManager.write_error(error_path, str(e))
            print(f"An error occured. Check {error_path}")
            close_everything()
            raise e


import os.path
from setuptools import setup

REQUIRES_PYTHON = ">=3.6.0"

HERE = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(HERE, "README.md"), encoding="utf-8") as fid:
    README = fid.read()

with open(os.path.join(HERE, "requirements.txt")) as fid:
    REQUIREMENTS = [req for req in fid.read().split("\n") if req]

from codequest22 import __version__

setup(
    name="codequest22",
    version=__version__,
    description="Game for Code Quest 2022",
    long_description=README,
    long_description_content_type="text/markdown",
    python_requires=REQUIRES_PYTHON,
    url="https://github.com/monash-programming-team/codequest22",
    author="Jackson Goerner, Ali Toosi",
    author_email="jgoerner@outlook.com, alitoosi137@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: pygame",
    ],
    packages=["codequest22"],
    include_package_data=True,
    install_requires=REQUIREMENTS,
    entry_points={
        "console_scripts": [
            "codequest22=codequest22:main",
        ]
    },
)




from contextlib import contextmanager
import time
import threading
import _thread

class TimeoutException(Exception):
    def __init__(self, msg=''):
        self.msg = msg

@contextmanager
def time_limit(seconds, msg=''):
    timer = threading.Timer(seconds, lambda: _thread.interrupt_main())
    timer.start()
    try:
        yield
    except KeyboardInterrupt:
        raise TimeoutException(f"Timed out for operation {msg}")
    finally:
        timer.cancel()

class BotRunner:
    """
    Wrapper for running bot commands.
    Allows a time limit to be set, as well as default functionality if the bot crashes or times out.
    """

    FUNCTIONS = [
        # name, and whether it is vital
        ("get_team_name", False),
        ("read_map", False),
        ("read_index", False),
        ("handle_events", False),
        ("handle_failed_requests", False),
    ]

    def __init__(self) -> None:
        self.registry = {}
        self.fallback = {}
        self.fallback["get_team_name"] = lambda : "Timed out"
        self.fallback["read_map"] = lambda _m, _m2: None
        self.fallback["read_index"] = lambda _i, _n: None
        self.fallback["handle_events"] = lambda _e: []
        self.fallback["handle_failed_requests"] = lambda _r: None

    def load_bot(self, bot_path):
        # Restart
        self.running = True
        self.over_elapsed = 0
        import importlib.util

        import os.path
        if os.path.isdir(bot_path):
            spec = importlib.util.spec_from_file_location("", os.path.join(bot_path, "main.py"))
            module = importlib.util.module_from_spec(spec)
            import sys
            sys.path.append(bot_path)
            spec.loader.exec_module(module)
        else:
            spec = importlib.util.spec_from_file_location("", bot_path)
            module = importlib.util.module_from_spec(spec)
            import sys
            sys.path.append(os.path.dirname(bot_path))
            spec.loader.exec_module(module)

        for key, vital in self.FUNCTIONS:
            if hasattr(module, key):
                self.registry[key] = getattr(module, key)
            elif vital:
                # We don't have a fallback
                raise ValueError(f"Bot {bot_path} must implement method {key}")
            else:
                print(f"Bot {bot_path} really should implement method {key}. We've provided a default.")

    def run_command(self, function_name, limit, *args, **kwargs):
        if not self.running:
            return self.fallback[function_name](*args, **kwargs)
        t0 = time.time()
        # Allow 3 times as much time to run.
        try:
            with time_limit(2 * limit, function_name):
                res = self.registry.get(function_name, self.fallback[function_name])(*args, **kwargs)
        except Exception as e:
            # If 2 or more times exceeded, assume failed.
            self.running = False
            res = self.run_command(function_name, limit, *args, **kwargs)
            if isinstance(e, TimeoutError):
                print("Bot spent longer than 2 times the time limit.")
            else:
                import traceback as tb

                error = "".join(tb.format_exception(None, e, e.__traceback__))
                print(f"Bot encountered Exception:\n{error}")
        else:
            elapsed = time.time() - t0
            if elapsed > limit:
                # If you spend 2 more seconds than you should be, then you die.
                self.over_elapsed += elapsed - limit
                if self.over_elapsed > 2:
                    self.running = False
                print("Bot spent more than 2 seconds total over limit.")
        return res



import math
import json
from codequest22.server.ant import AntTypes
from codequest22.server.events import AttackEvent, DepositEvent, DieEvent, MoveEvent, SettlerScoreEvent, SpawnEvent
from codequest22.visual.game_input import _rec_finish_dict

class GameSummariser:

    @classmethod
    def get_results(cls, replay_path: str) -> dict:
        with open(replay_path, "r") as f:
            replay_obj = [_rec_finish_dict(json.loads(line)) for line in f.readlines()]
        return {
            "ants": {
                "worker": cls.get_worker_stats(replay_obj),
                "fighter": cls.get_fighter_stats(replay_obj),
                "settler": cls.get_settler_stats(replay_obj),
            },
            "misc": cls.get_misc_stats(replay_obj)
        }
    
    @classmethod
    def get_worker_stats(cls, replay_obj):
        data = None
        worker_ids = None
        for line in replay_obj:
            if line["type"] == "player_data":
                data = [{
                    "spawned": 0,
                    "deposits": 0,
                    "total_deposited": 0,
                    "hp_died": 0,
                    "lifetime_died": 0,
                } for _ in range(len(line["names"]))]
                worker_ids = [
                    set() for _ in range(len(line["names"]))
                ]
            elif line["type"] == "tick":
                for ev in line["events"]:
                    if isinstance(ev, SpawnEvent) and ev.ant_type == AntTypes.WORKER:
                        data[ev.player_index]["spawned"] += 1
                        worker_ids[ev.player_index].add(ev.ant_id)
                    elif isinstance(ev, DepositEvent) and ev.ant_id in worker_ids[ev.player_index]:
                        data[ev.player_index]["deposits"] += 1
                        data[ev.player_index]["total_deposited"] += ev.energy_amount
                    elif isinstance(ev, DieEvent) and ev.ant_id in worker_ids[ev.player_index]:
                        if ev.old_age:
                            data[ev.player_index]["lifetime_died"] += 1
                        else:
                            data[ev.player_index]["hp_died"] += 1
        return data

    @classmethod
    def get_fighter_stats(cls, replay_obj):
        data = None
        fighter_ids = None
        for line in replay_obj:
            if line["type"] == "player_data":
                data = [{
                    "spawned": 0,
                    "kills": 0,
                    "hp_died": 0,
                    "lifetime_died": 0,
                } for _ in range(len(line["names"]))]
                fighter_ids = [
                    set() for _ in range(len(line["names"]))
                ]
            elif line["type"] == "tick":
                for ev in line["events"]:
                    if isinstance(ev, SpawnEvent) and ev.ant_type == AntTypes.FIGHTER:
                        data[ev.player_index]["spawned"] += 1
                        fighter_ids[ev.player_index].add(ev.ant_id)
                    elif isinstance(ev, AttackEvent) and ev.attacker_id in fighter_ids[ev.attacker_index]:
                        if ev.defender_hp <= 0:
                            data[ev.attacker_index]["kills"] += 1
                    elif isinstance(ev, DieEvent) and ev.ant_id in fighter_ids[ev.player_index]:
                        if ev.old_age:
                            data[ev.player_index]["lifetime_died"] += 1
                        else:
                            data[ev.player_index]["hp_died"] += 1
        return data
    
    @classmethod
    def get_settler_stats(cls, replay_obj):
        data = None
        settler_ids = None
        for line in replay_obj:
            if line["type"] == "player_data":
                data = [{
                    "spawned": 0,
                    "hp_died": 0,
                    "lifetime_died": 0,
                    "points": 0,
                } for _ in range(len(line["names"]))]
                settler_ids = [
                    set() for _ in range(len(line["names"]))
                ]
            elif line["type"] == "tick":
                for ev in line["events"]:
                    if isinstance(ev, SpawnEvent) and ev.ant_type == AntTypes.SETTLER:
                        data[ev.player_index]["spawned"] += 1
                        settler_ids[ev.player_index].add(ev.ant_id)
                    elif isinstance(ev, SettlerScoreEvent) and ev.ant_id in settler_ids[ev.player_index]:
                        data[ev.player_index]["points"] += ev.score_amount
                    elif isinstance(ev, DieEvent) and ev.ant_id in settler_ids[ev.player_index]:
                        if ev.old_age:
                            data[ev.player_index]["lifetime_died"] += 1
                        else:
                            data[ev.player_index]["hp_died"] += 1
        return data

    @classmethod
    def get_misc_stats(cls, replay_obj):
        # IDEAS
        # Most aggressive - Highest ratio of fighter ants to other spawns
        # Marathon Runner - Longest average distance travelled per ant
        # Waiting for the right moment - Highest stored energy
        # Stayin Alive - Had the highest ratio of ants dying by age compared to HP loss.
        misc = {}
        # Most aggressive.
        fighter_spawned = []
        other_spawned = []
        for line in replay_obj:
            if line["type"] == "player_data":
                fighter_spawned = [0] * len(line["names"])
                other_spawned = [0] * len(line["names"])
            elif line["type"] == "tick":
                for ev in line["events"]:
                    if isinstance(ev, SpawnEvent):
                        if ev.ant_type == AntTypes.FIGHTER:
                            fighter_spawned[ev.player_index] += 1
                        else:
                            other_spawned[ev.player_index] += 1
        misc["aggressive"] = {
            "fighter": fighter_spawned,
            "other": other_spawned,
            "ratio": [(f/(f+o) if f+o > 0 else 0) for f, o in zip(fighter_spawned, other_spawned)]
        }
        # Marathon runner
        distance_travelled = []
        cur_positions = []
        total_spawns = []
        for line in replay_obj:
            if line["type"] == "player_data":
                distance_travelled = [0] * len(line["names"])
                total_spawns = [0] * len(line["names"])
                cur_positions = [{} for _ in range(len(line["names"]))]
            elif line["type"] == "tick":
                for ev in line["events"]:
                    if isinstance(ev, SpawnEvent):
                        total_spawns[ev.player_index] += 1
                        cur_positions[ev.player_index][ev.ant_id] = ev.position
                    elif isinstance(ev, MoveEvent):
                        old_pos = cur_positions[ev.player_index][ev.ant_id]
                        distance_travelled[ev.player_index] += math.sqrt((ev.position[0] - old_pos[0])**2 + (ev.position[1] - old_pos[1])**2)
        misc["marathon"] = {
            "distance": distance_travelled,
            "spawns": total_spawns,
            "avg": [(d/s if s > 0 else 0) for d, s in zip(distance_travelled, total_spawns)]
        }
        # Waiting for the right moment
        cur_max = []
        for line in replay_obj:
            if line["type"] == "player_data":
                cur_max = [0] * len(line["names"])
            elif line["type"] == "tick":
                for ev in line["events"]:
                    if isinstance(ev, DepositEvent):
                        # Deposit is the only way to increase energy
                        cur_max[ev.player_index] = max(cur_max[ev.player_index], ev.cur_energy)
        misc["waiting"] = {
            "max_energy": cur_max
        }
        # Stayin Alive
        hp_loss = []
        age = []
        for line in replay_obj:
            if line["type"] == "player_data":
                hp_loss = [0] * len(line["names"])
                age = [0] * len(line["names"])
            elif line["type"] == "tick":
                for ev in line["events"]:
                    if isinstance(ev, DieEvent):
                        if ev.old_age:
                            age[ev.player_index] += 1
                        else:
                            hp_loss[ev.player_index] += 1
        misc["alive"] = {
            "hp_loss": hp_loss,
            "age": age,
            "pct": [(a/(a+h) if a+h > 0 else 0) for a, h in zip(age, hp_loss)]
        }
        return misc

if __name__ == "__main__":
    print(GameSummariser.get_results("replay.txt"))



import os
GRACE_PERIOD = int(os.environ.get("HILL_GRACE_PERIOD", "100"))
NUM_ACTIVATIONS = int(os.environ.get("HILL_NUM_ACTIVATIONS", "2"))
MIN_ZONE_TIME = int(os.environ.get("HILL_MIN_ZONE_TIME", "60"))
MAX_ZONE_TIME = int(os.environ.get("HILL_MAX_ZONE_TIME", "80"))
MIN_WAIT_TIME = int(os.environ.get("HILL_MIN_WAIT_TIME", "20"))



import os
DELAY = int(os.environ.get("ENERGY_DELAY", "2"))
PER_TICK = int(os.environ.get("ENERGY_PER_TICK", "1"))
GRACE_PERIOD = int(os.environ.get("ENERGY_GRACE_PERIOD", "100"))
MIN_OVERCHARGE_TIME = int(os.environ.get("ENERGY_MIN_OVERCHARGE_TIME", "50"))
MAX_OVERCHARGE_TIME = int(os.environ.get("ENERGY_MAX_OVERCHARGE_TIME", "80"))
MIN_WAIT_TIME = int(os.environ.get("ENERGY_MIN_WAIT_TIME", "50"))
NUM_ACTIVATIONS = int(os.environ.get("ENERGY_NUM_ACTIVATIONS", "3"))